// fuzz.js 3.0


// fuzz.js provides the following features for fuzzers that use it:
// * Seeded pseudorandom numbers
// * Chunking of steps for speed
// * Recording and playing, to make reduction possible.  (Even takes chunking into account most of the time.)
// * Bookmarklet mode
// * Things to aid automated reduction with Lithium (DD* markers, goQuitApplication)
// * Things to aid automated testing with Spider
// * Pausing (Esc pauses, Shift+Esc unpauses)
// * And, by being a "framework", consistency between fuzzers and hopefully easier maintainance.

// Rules for fuzzers:
// * Defined the following variables: fuzzerName, recordStrategy.
// * makeCommand should try to be free of side-effects.
// * doCommand should be free of randomness.
// * init should call startFuzzing (either immediately, or if needed, using timeout).
// * Use "doc", not "document" (for spider).
// * Use "win", not "window" (for spider).
//   * window.status is probably an exception.
//   * Be careful when using "if (window.foo)" to test whether variables exist.

// How to reduce a crash found with a fuzzing bookmarklet:
// 1. Save the page on which you found the crash.  (If it's an HTML page, use the "HTML only" saving mode.)
// 2. Paste the chunk of the fuzz script page containing the two <script> elements (one with src="fuzz.src", one inline) into the saved page.
//      For anything that has a document.body (HTML or XHTML), paste it in <head> (below meta charset, if present -- bug 322181).
//      For things like SVG, paste it so that it's the first child of the root, with no whitespace before it.
// 3. Load the page.
// 4. A dialog should appear asking for fuzzing parameters.  Give it the same parameters you gave the bookmarklet, and verify that it still crashes.
// 5. Do step 4 again, but add the fifth (maxSteps) parameter (overshooting a bit is ok) and make the sixth paramter 1 to tell it to record.
// 6. For some fuzzers, you'll see a big blog of pastable code.  Paste it into the testcase, replacing the lines "var fuzzSettings = null;" and "var fuzzCommmands = null;".  (For others fuzzers, you'll have to grep console output for origCount to create the fuzzCommands array; this process isn't as user-friendly.)
// 7. Verify that it still crashes.
// 8. Reduce the testcase, usually starting with the fuzzCommands array.  This is usually the most time-consuming part of a reduction, but Lithium (https://bugzilla.mozilla.org/show_bug.cgi?id=329066) can often automate it.
// 9. Reduce the rest of the testcase manually.


var dumpEachCommand = false; // Can be set to true by using the "Record as it goes" reduction strategy.  Also, some fuzzers always set this to true.
var dumpEachSeed = false; // Set to true if makeCommand has side effects, such as crashing, so you have to reduce "the hard way".
var gPageCompleted = false;

// like spidermonkey's uneval, but cross-browser, and only for strings and simple things like numbers.
function simpleSource(s)
{
  if (typeof s == "string")
    return "\"" + s.replace(/\\/g, "\\\\").replace(/\"/g, "\\\"") + "\"";
  else
    return "" + s; // hope this is right ;)  should work for numbers.
}

function fuzzOnload(ev)
{
  window.removeEventListener("load", fuzzOnload, false); // don't init twice when using a bookmarklet

  initFuzzerGeneral();

  // initFuzzerSpecific() is expected to call startFuzzing.  Most fuzzers call it immediately but some use a timeout or callback.
  initFuzzerSpecific();
}

function initFuzzerGeneral()
{
  obtainSettings();

  dumpln(fuzzerName + " running on " + win.location.href);
  dumpln("var fuzzSettings = [" + fuzzSettings + "];");

  rnd = randomizer(fuzzSeed);
}


function startFuzzing()
{
  if (window.fuzzCommands) {
    // Already recorded, so we should play.
    count = 0;
    if (recordStrategy == "Special") {
      playFunsSpecial();
    } else {
      playFunsChunk();
    }
  }
  else {
    // Not already recorded
    if (recordMode) {
      recordSomehow();
    } else {
      immedChunk(numImmediate);
    }
  }
}

function recordSomehow()
{
  switch(recordStrategy) {
    case "Record without doing": 
      recordFuns();
      break;

    case "Record as it goes":
      alert("This fuzzer records as it goes.  If you encounter a crash or other bug you want to reduce, grep console output for lines containing origCount and paste that in as a fuzzCommands array.");
      dumpEachCommand = true;
      immedChunk(numImmediate);
      break;

    case "Special":
      recordFunsSpecial(); // defined by the fuzzer
      break;

    case "None":
    default:
      alert("This fuzzer doesn't seem to support recording.");
  }
}


var immedCount = 0;

function immedChunk(changes)
{
  if (gPageCompleted) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (done!)";
  }
  else if (pauseFuzzing) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (paused)";
  }
  else {   
    // Not paused or finished.
    
    if (immedCount > maxSteps) {
      gPageCompleted = true; // Tell Spider we're done.
      pauseFuzzing = true; // Don't keep printing useless stuff every 100 steps, don't keep trying to load the next page in the scan.
    } 
    else {

      for (var i = 0; i < changes; ++i) {
        ++immedCount;
        if (!pauseFuzzing && immedCount <= maxSteps)
          immedStep();
      }

      if (dumpEachCommand) {
        dumpln("  { origCount: " + immedCount + ", rest: true },");
      }

    }
  }

  // gPageCompleted may have been set before immedChunk() was called, or
  // as a result of immedStep(), or by code *in* immedChunk!?
  if (gPageCompleted) {
     if (getURLParam("scan")) {
      // Take advantage of seed being the first parameter to replace it easily.
      location.search = location.search.replace("fuzz=" + fuzzSeed, "fuzz=" + (fuzzSeed + 1));
    }
  }
  else if (stepsPerInterval != 0)
    setTimeout(function() { immedChunk(stepsPerInterval); }, interval);
}





function immedStep()
{
  if (dumpEachSeed) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { rnd.setSeed(" + rnd.getSeed() + "); immedStep(); } },");
  }
  
  var x = makeCommand();
  window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount;
  // randomJS.collectGarbage(); // Whee!  I put this right after "window.status" because that's what I see in the crashes...

  if (dumpEachCommand) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { " + x + " } },");
  } else {
    // Dump stuff occasionally so that if we hit a crash, we have a hint as to where we hit it.
    if (immedCount % (100) == 0)
      dumpln(fuzzerName + ": " + immedCount);
  }


  try {
    doCommand(x);
    // how will eval work? will "this" be the fuzzer?
  }
  catch(e) { 
    // Should this report the error somehow?
  }
}



/************
 * SETTINGS *
 ************/
 
var fuzzSeed = 0;
var numImmediate = 0;
var stepsPerInterval = 100;
var interval = 400;
var maxSteps = Infinity;
var recordMode = 0;


function obtainSettings()
{

  // if (window.fuzzSettings === undefined) alert("No settings variable declaration? No fuzzer?");
  // Let's just allow this to hit an error, instead of putting up an annoying alert whenever a fuzzer has a typo or we're reducing like crazy but keeping goQuitApplication.
  /*
  if (fuzzSettings == null)
    fuzzSettings = getURLParam("fuzz");

  if (fuzzSettings == null)
    fuzzSettings = prompt(fuzzerName + " settings\n\nSeed\nNumber of changes in first chunk\nNumber of changes in each subsequent chunk\nInterval between chunks in milliseconds\nTotal number of changes (or 0 for unlimited)\n1 to record, 0 to go", "0, 0, 100, 400, 0, 0");

  if (fuzzSettings == null) {
    fuzzSeed = 0; // for playing around with javascript: URLs, accessing allNodes, etc.
    numImmediate = stepsPerInterval = maxSteps = 0;
    interval = 1000;
    // throw "User hit cancel!";
    return;
  }

  if (typeof fuzzSettings == "string") {
    fuzzSettings = fuzzSettings.replace(/;/g, ","); // allow semicolons in place of commas (makes Talkback's URL field happier)
    fuzzSettings = eval("[" + fuzzSettings + "]");
  }

  if (fuzzSettings[0]) fuzzSeed         = fuzzSettings[0];
  if (fuzzSettings[1]) numImmediate     = fuzzSettings[1];
  if (fuzzSettings[2]) stepsPerInterval = fuzzSettings[2];
  if (fuzzSettings[3]) interval         = fuzzSettings[3];
  if (fuzzSettings[4]) maxSteps         = fuzzSettings[4]; // if it's 0, it remains Infinity!
  if (fuzzSettings[5]) recordMode       = fuzzSettings[5];
  // "force quit" option should be here, not in the pasted fuzzCommands array!
//0, 0, 100, 400, 0, 0
*/


 fuzzSeed         = 0;
numImmediate     = 0;
 stepsPerInterval = 100;
 interval         = 400;
 maxSteps         = 0; // if it's 0, it remains Infinity!
  recordMode       = 0;
}

function getURLParam(p)
{
  var s = win.location.search + "&"; 
  var r = new RegExp("[?&]" + p + "\\=([^&]*)\\&");
  if(r.exec(s) == null)
    return null;
  var raw = RegExp.$1;
  return unescape(raw.replace(/\+/g," "));
}


/***********
 * PAUSING *
 ***********/
 
var pauseFuzzing = false;

function escPause(ev)
{
  if (ev.keyCode == 27) { // esc
    if (!ev.shiftKey) {
      dumpln("Pausing fuzzing.");
      pauseFuzzing = true;
    } else {
      // Shift+Esc (works in Firefox but not in Safari)
      dumpln("Resuming fuzzing.");
      pauseFuzzing = false;
    }
  }
}

window.addEventListener("keypress", escPause, false);




/*********************************
 * SIMPLE RECORDING AND PLAYBACK *
 *********************************/
 
// This section is for when recordStrategy == "Record without doing".

function playFuns()
{
  var command = fuzzCommands[count];

  ++count;

  if (command == undefined)
    return false;
  
  var message = fuzzerName + " (playing): origCount " + command.origCount;
  window.status = message;

  // Also dump sometimes, in case Firefox has stopped painting, or in case we want to see info related to a crash *before reducing*
  if (command.origCount % (100) == 0 && !command.rest)
    dumpln(fuzzerName + " (playing): origCount " + command.origCount);

  if (command.fun) {
    try { 
      command.fun(); 
    } catch(e) {
      var errorAsString = "";
      try {
        errorAsString = "" + e;
      } catch(e2) { 
        errorAsString = "[cannot convert this exception to string!]";
      }
      window.status += " threw: " + errorAsString; 
    }
  }

  return !(command.rest);
}



function playFunsChunk()
{
  while(playFuns())
    ;

  // Set a timer for the next chunk
  setTimeout(playFunsChunk, interval);
}


function recordFuns()
{
  // This is nice in that it avoids verbose output...
  var fuzzCommands = [];

  var output =
     "var fuzzSettings = [" + fuzzSettings + "];\n" +
     "var fuzzCommands = [\n" +
     "// DD" + "BEGIN\n";
     
  for (var count = 1; count <= maxSteps; ++count)
  {
    var command = makeCommand();
    output += "  { origCount: " + count + ", fun: function() { " + command + " } },\n";

    var countAfterImmed = count - numImmediate;
    if((countAfterImmed >= 0) && (countAfterImmed % stepsPerInterval == 0)) {
      output += "  { origCount: " + count + ", rest: true },\n";
    }
  }
  
  output += 
    "// DD" + "END\n" +
    "];";

  giveToCopy("// Paste this into the script, replacing the first two lines:\n\n" + output);
  
  // Verbosity of output is due to in part to
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322175 and
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322177

}




/***************************
 * REPRODUCIBLE RANDOMNESS *
 ***************************/
 

// This random number generator obviously repeats pretty often :(  Are there better RNGs available in (or for) JavaScript?

// Given a seed, returns a function that generates random integers.
// Based on the Central Randomizer 1.3 (C) 1997 by Paul Houle (paul@honeylocust.com)
// See: http://www.honeylocust.com/javascript/randomizer.html
function randomizer(seed)
{
  
  // Returns a somewhat random real number between 0 (inclusive) and 1 (not inclusive), like Math.random() but with a seed I control.
  // Modulus is in parens to work around a Crunchinator bug.
  function rndReal() {
	seed = (seed*9301 + 49297) % (233280);
	return seed/(233280.0);
  };

  // Returns a somewhat random integer in the range {0, 1, 2, ..., number-1}
  function rndInt(number) {
	return Math.floor(rndReal()*number);
  };
  
  // for hard reductions only
  rndInt.setSeed = function(s) { seed = s; };
  rndInt.getSeed = function() { return seed; };
  
  return rndInt;
}


// Based on http://archives.hwg.org/hwg-basics/002d01c11d20$c3dec4c0$0200000a@matthew
// Nope, this is even worse (e.g. with seed = 4)
/*
function randomizer(seed)
{
  function rndInt(number)
  {
    seed = (0x015a4e35 * seed) % (0x7fffffff);
    return (seed >> 16) % number;
  }
  
  return rndInt;
}
*/


function rndElt(a) { return a[rnd(a.length)]; }



/**********
 * OUTPUT *
 **********/
 
function dumpln (s)
{
  if (window.dump) 
    dump(s + "\n");
}


function giveToCopy(s)
{
  // This can crash on Linux (due to a long string in the address bar?), so we should have an alternate strategy...
  location.href = "data:text/plain;charset=UTF-8," + encodeURIComponent(s);
  
  // this is nice, but only works is HTML (see bug 332848): 
  //document.open("text/plain");
  //document.write(s);
  
  // other strategies: add an <html:pre> to the end of the document, ...
  
  // When anything else would be too painful
  //dumpln(s);
  //alert("Dumped");
}




/**************
 * FORCE QUIT *
 **************/

// From quit.js, which Bob Clary extracted from mozilla/toolkit/content

function canQuitApplication()
{
  var os = Components.classes["@mozilla.org/observer-service;1"]
    .getService(Components.interfaces.nsIObserverService);
  if (!os) 
  {
    return true;
  }

  try 
 {
    var cancelQuit = Components.classes["@mozilla.org/supports-PRBool;1"]
      .createInstance(Components.interfaces.nsISupportsPRBool);
    os.notifyObservers(cancelQuit, "quit-application-requested", null);
  
    // Something aborted the quit process. 
    if (cancelQuit.data)
    {
      return false;
    }
  }
  catch (ex) 
  {
  }
  os.notifyObservers(null, "quit-application-granted", null);
  return true;
}

function goQuitApplication()
{
  var privs = 'UniversalPreferencesRead UniversalPreferencesWrite ' +
    'UniversalXPConnect';

  try
  {
    netscape.security.PrivilegeManager.enablePrivilege(privs);
  }
  catch(ex)
  {
    throw('goQuitApplication: privilege failure ' + ex);
  }

  if (!canQuitApplication())
  {
    return false;
  }

  var kAppStartup = '@mozilla.org/toolkit/app-startup;1';
  var kAppShell   = '@mozilla.org/appshell/appShellService;1';
  var   appService;
  var   forceQuit;

  if (kAppStartup in Components.classes)
  {
    appService = Components.classes[kAppStartup].
      getService(Components.interfaces.nsIAppStartup);
    forceQuit  = Components.interfaces.nsIAppStartup.eForceQuit;

  }
  else if (kAppShell in Components.classes)
  {
    appService = Components.classes[kAppShell].
      getService(Components.interfaces.nsIAppShellService);
    forceQuit = Components.interfaces.nsIAppShellService.eForceQuit;
  }
  else
  {
    throw 'goQuitApplication: no AppStartup/appShell';
  }

  var windowManager = Components.
    classes['@mozilla.org/appshell/window-mediator;1'].getService();

  var windowManagerInterface = windowManager.
    QueryInterface(Components.interfaces.nsIWindowMediator);

  var enumerator = windowManagerInterface.getEnumerator(null);

  while (enumerator.hasMoreElements())
  {
    var domWindow = enumerator.getNext();
    if (("tryToClose" in domWindow) && !domWindow.tryToClose())
    {
      return false;
    }
    domWindow.close();
  }

  try
  {
    appService.quit(forceQuit);
  }
  catch(ex)
  {
    throw('goQuitApplication: ' + ex);
  }

  return true;
}



/***************
 * OTHER MODES *  
 ***************/


var win = (window.gSpider ? gSpider.mDocument.defaultView : window); // For spider.  I hope this is right (and safe).
var doc = win.document;


// Non-bookmarklet
//window.addEventListener("load", fuzzOnload, false); // !!!!

// Bookmarklet
// fuzzOnload();



/********************
 * COMMON DOM STUFF *
 ********************/
 
function addDOMNodes(n, enterIFrames, enterSVG, enterXBL) { 
  var i,c;

  if (n.tagName && n.tagName.toLowerCase().indexOf("script" != -1) && n.getAttribute("id") && n.getAttribute("id").indexOf("fuzz") != -1) {
    if (n.previousSibling && n.previousSibling.nodeType == Node.TEXT_NODE)
      alert("Warning: fuzz script found with a text (whitespace?) node before it.  This could affect node counts.");
    return;
  }

  if (n.tagName && n.tagName.toLowerCase() == "frameset") {
    gPageCompleted = true;
    throw "Avoiding bug 324318!";
  }
  

  // Add this node
  allNodes.push(n); 
  
  // Look for children
  for (i=0;c=n.childNodes[i];++i) {
    //try {
      addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
    //} catch(e) { /* Why would this fail? */ }
  }
  
  // If this is a frame, look inside it.
  if(enterIFrames) {
    try {
      if(n.contentDocument)
        try {
          addDOMNodes(n.contentDocument.documentElement, enterIFrames, enterSVG, enterXBL); 
        } catch(ex) { /* third-party iframe */ }
    } catch(ex) { /* avoid bug 310994 */ }
  }
  
  // If this is an <embed> referencing an SVG image, look inside it for SVG nodes.
  if (enterSVG) {
    if (n.getSVGDocument) { 
      try{
        if (n.getSVGDocument())
          addDOMNodes(n.getSVGDocument(), enterIFrames, enterSVG, enterXBL);
      } catch(ex) { }
    }
  }

  // Look for anonymous children (XBL, at least)
  // try/catch in case browser doesn't support getAnonymousNodes, or in case Firefox decides to throw a FAILURE error (!?).
  if (enterXBL) {
    try { 
      if (n.nodeType==1 && doc.getAnonymousNodes(n)) {
        for(i=0;c=doc.getAnonymousNodes(n)[i];++i) {
          if (!this.anonymousCount)
            anonymousCount = 0;
          ++anonymousCount;
          addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
        }
      }
    } catch (ex) { } 
  }
}

// === Paste output from first run here, replacing the following two lines: ===

var fuzzSettings = null;
var fuzzCommands = null;

// === End pasteover ===


var fuzzerName = "HTML Element and Attribute fuzzer 3.0";
var recordStrategy = "Record as it goes";
var suggestedMaxSteps = 20000;

// dumpEachCommand = true;


var allNodes;
var fuzzRoot;

var HTML_NS = "http://www.w3.org/1999/xhtml";

// dumpEachCommand = true;


function aC(q1, q2) { q1.appendChild(q2); }
function iB(q1, q2) { q1.parentNode.insertBefore(q2, q1); }
function rM(q1) { q1.parentNode.removeChild(q1); } // unused, but can be useful while narrowing testcases
// other things to try: clone instead of move, ...

function makeCommand()
{
  var n1index = rnd(allNodes.length), n2index = rnd(allNodes.length);
  
  var n1 = allNodes[n1index];
  var n2 = allNodes[n2index];
  
  var commandn1 = "allNodes[" + n1index + "]";
  var commandn2 = "allNodes[" + n2index + "]";
  var command;
  
  while(1){
  switch (rnd(9)) {
  case 0:
  case 1:
  case 2:
  
    // Change an attribute on n1, based on n1's tagName.
    
    if (!n1.tagName)
      break;
      
    var attrs = rnd(2) ? elements[n1.tagName] : commonAttributes;
    var attr = attrs ? rndElt[attrs] : null;
    if (!attr || rnd(5) == 1) {
      attr = totallyRandomAttr();
      // dumpln("*** " + attr);
    }
      
    var value = randomThing(rnd(5) == 1 ? attributes[totallyRandomAttr()] : attributes[attr]);
    
    if (attr) {
      if (rnd(2))
        // Set it as a property.
        return commandn1 + "['" + attr + "'] = " + simpleSource(value) + ";";
      else 
        // Set it as an attribute.
        return commandn1 + ".setAttribute('" + attr + "', " + simpleSource(value) + ");";
    }
      
    break;
    
  case 3:
    // Remove an attribute on n1, based on n1's current attribute list.
    
    var attrs = n1.attributes;
    if (attrs && attrs.length) {
      var attr = rndElt(attrs).name;
      return commandn1 + ".removeAttribute('" + attr + "');";
    }
    
    break;
    
  case 4:
    // Create a new element!
    // Set some attributes on it?  Nah.
    // Place it randomly, but only rarely place it outside of the document, because that's usually boring.
    
    var tag = rndElt(elementList);
    var newIndex = allNodes.length;
    

    var createness;
    if ((tag == "foo") != (rnd(10) == 1))
      createness = "var ne = document.createElementNS('bar', '" + tag + "'); allNodes[" + newIndex + "] = ne; ";
    else
      createness = "var ne = document.createElementNS(HTML_NS, '" + tag + "'); allNodes[" + newIndex + "] = ne; ";
    
    
    function isInDoc(n) { if (!n) return false; if (n.nodeType == Node.DOCUMENT_NODE) return true; return isInDoc(n.parentNode); }

    if (isInDoc(n1) != (rnd(10) == 1))
      return createness + commandn1 + ".appendChild(ne); ";
    else
      return createness + "fuzzRoot.appendChild(ne); ";

    break;

  case 5:
  case 6:
  case 7:
    // Move n2 in some way..
    
    if(n2index > 1) { // This fuzzer avoids screwing with the root node.
  
      if (rnd(20) == 1)
        // Don't do this too often -- it tends to make things less interesting.
        return "rM(" + commandn2 + ");";
      
      if (rnd(9) == 3)
        // Don't do this too often either -- it hurts reduction and it doesn't buy us much
        return "iB(" + commandn1 + ", " + commandn2 + ");";
        
      // Do this a lot ;)
      return "aC(" + commandn1 + ", " + commandn2 + ");";
    }
    
    break;
    
  case 8:
    // Add some text -- this is partly for visual effect rather than finding more bugs.
  
    if (rnd(4) == 1)
      return "aC(" + commandn1 + ", allNodes[" + allNodes.length + "] = document.createTextNode(''));";
    else if (rnd(3) == 1)
      return "aC(" + commandn1 + ", allNodes[" + allNodes.length + "] = document.createTextNode('" + decodeURIComponent("%28%D7%A2%D6%B4%D7%91%D6%B0%D7%A8%D6%B4%D7%99%D7%AA") + "'));"; // hebrew
    else if (rnd(4) == 1)
      return "aC(" + commandn1 + ", allNodes[" + allNodes.length + "] = document.createTextNode('" + decodeURIComponent("%D8%A7%D9%84%D8%B9%D8%B1%D8%A8%D9%8A") + "'));"; // arabic?
    else
      return "aC(" + commandn1 + ", allNodes[" + allNodes.length + "] = document.createTextNode('<" + n1.tagName + ">'));";

  }
  }
  
}

function doCommand(x) { 

try {
  eval(x); 
} catch(e) {
//  alert(e);
//  gPageCompleted = true;
}

}




function resetNodeList() {
  allNodes = [];
  fuzzRoot = doc.body || doc.documentElement;
  addDOMNodes(fuzzRoot);
}


function initFuzzerSpecific() { 
  if (document.body)
    document.body.style.width = "300px"; // encourage wrapping
  resetNodeList();
  startFuzzing();
}



function randomThing(obj)
{
  if (obj == null)
    return null;
    
  if (typeof obj == "function")
    return obj();
    
  if (typeof obj == "string")
    return obj; // not to be treated as an array :)

  if (obj.length) { // non-empty array
    var x = rndElt(obj);
    //dumpln(x);
    return randomThing(x)
  }

  if (obj.length === 0)
    return null;

  // Otherwise, it should be a number, etc.
  return obj;
}



var names =    [  "a",  "b",  "c",  "d",  "e",  "f",  "g",  "1" ];
var namerefs = [ "#a", "#b", "#c", "#d", "#e", "#f", "#g", "#1" ];
var charsets = [ "UTF-8", "ISO-8859-1" ];
var URIs = ["http://www.squarefree.com/", "http://www.squarefree.com/feed", "#a", "#b", "data:text/html,", /* "javascript:alert(Components.classes)", "javascript:5555", "javascript:'QQQQ' + String.fromCharCode(0) + 'UUUU'" -- avoid bug 343940 */,  "http://www.mozilla.org/status/tree.gif"];
var numbers = [-2, -1, 0, "-0", "+0", 1, "+1", 2, 3, 4, 5, 6, 7, 8, 100];
function numbersWithUnits() { return rndElt(numbers) + rndElt(["px", "em", "pt", "%", ""]) }
var relations = ["alternate"];
var mimeTypes = ["text/html", "text/plain", "text/css", "text/javascript", "application/rss+xml"];
var texts = ["", "foo", "bar", "baz", "zap", "Abbey\nRoad", 'QQQQ' + String.fromCharCode(0) + 'UUUU'];
var colors = ["lime", "red", "blue", "#CCC", "#777777"];

function longThing()
{
  // a string of all A's in a buffer overflow has the nice property
  // of a guaranteed crash on some architectures, IIRC
  
  var x = "A"; 
  for (var i = 0; i < 14; ++i)
    x = x + x;
  return x; 
}


var commonAttributes = ["id", "name", "width", "height"];

var attributes = {
"randomcrap": ["", "\0", longThing],
"id": names,
"name": names,
"for": names, // does not use #.  what about <script for> -- is that different?
"usemap": namerefs, // uses #
"href": [namerefs, URIs],
"type": [
  mimeTypes /* link element*/, 
  ["text", "password", "file", "radio", "checkbox", "submit", "reset", "image", "button"],
  ["i","1","A","I","a"] // list types in HTML 3.2
  ],
"enctype": mimeTypes,
"accept": mimeTypes,
"accept-charset": charsets,
"charset": charsets,
"action": URIs,
"method": ["GET", "POST"],
"maxlength": numbers,
"tabindex": numbers,
"width": [numbers, numbersWithUnits],
"height": [numbers, numbersWithUnits],
// "dir": ["rtl", "ltr"], // commented out to avoid triggering what i think is bug 334147
"checked": "checked",
"disabled": "disabled",
"readonly": "readonly",
"size": numbers,
"maxlength": numbers,
"src": URIs,
"alt": texts,
"ismap": "ismap", // ?
"accesskey": names, // hack
"multiple": "multiple", // ?
"selected": "selected",
"label": texts,
"rows": numbers,
"cols": numbers,
"rel": relations,
"rev": relations,

// Image maps
"shape": "rect",
"coords": "5,10,15,20",
"nohref": "nohref", // ?

"longdesc": URIs,

"object": ["declare", "classid", "codebase", "data", "type", "codetype", "archive", "standby", "height", "width", "usermap", "name", "tabindex"],
"param": ["id", "name", "value", "valuetype", "type"],

"pluginspage": URIs,
"cite": URIs,
"datetime": "2006-06-04", // ?




// "meta": "http-equiv", "name", "content", "scheme"


"media": [], // ?
"defer": "defer",
"event": [],



"align": ["left", "center", "right", "justify", ":", // HTML 4 for tables (colon is for "char" in the HTML 4 spec (?))
          "top", "middle", "bottom", "left", "right" ], // HTML 3.2 for images 

"vspace": numbers,
"hspace": numbers,

"bgcolor": colors,
"text": colors,
"link": colors,
"vlink": colors,
"alink": colors,

"color": colors,

"background": URIs,

"noshade": "noshade", // ?


// Tables -- be nice to them for now; stirtable takes care of most of it.
/*
"valign": ["top", "middle", "bottom", "baseline"],
"border": numbers,
"cellspacing": numbers,
"cellpadding": numbers,
"summary": texts,
"span": numbers,
"rowspan": numbers,
"colspan": numbers,
"scope": ["row", "col", "rowgroup", "colgroup"],
"headers": names, // really multiple names
"axis": names, // ?

// Obscure stuff from the TABLES section of http://www.w3.org/TR/REC-html40/sgml/dtd.html
"frame": ["void", "above", "below", "hsides", "lhs", "rhs", "vsides", "box", "border"],
"rules": ["none", "groups", "rows", "cols", "all"]
*/

// marquee
"behavior": ["scroll", "slide", "alternate"],
"direction": ["left", "right", "down", "up"],
"loop": [-1, 0, 1, 2, numbers],
"scrollamount": numbers,
"scrolldelay": numbers,
"truespeed": numbers

};

// attributes is a hash, but sometimes we want to treat it as an array-like thing.  so....
attributeList = [];
for (var prop in attributes)
  attributeList.push(prop);

function totallyRandomAttr() { 
  // Emphasize basic stuff. 
  if (rnd(3)==1) 
    return rndElt(["id", "name"]);
  return rndElt(attributeList); 
}



// Text markup

var elements = 
{
"tt": [],
"i": [],
"b": [],
"big": [],
"small": [],
"em": [],
"strong": [],
"dfn": [],
"code": [],
"samp": [],
"kbd": [],
"var": [],
"cite": [],
"abbr": [],
"acronym": [],
"br": [],
"map": [],
"q": ["cite"],
"sub": [],
"sup": [],
"span": [],
"bdo": ["dir"], // but everything can have dir, but it's required for bdo
"input": ["type", "name", "value", "checked", "disabled", "readonly", "size", "maxlength", "src", "alt", "usemap", "ismap", "tabindex", "accesskey", "accept"],
"select": ["name", "size", "multiple", "disabled", "readonly", "tabindex"],
"option": ["selected", "disabled", "label", "value"],
"optgroup": ["label", "disabled"],
"textarea": ["name", "rows", "cols", "disabled", "readonly", "tabindex", "accesskey"],
"button": ["name", "value", "type", "disabled", "tabindex", "accesskey"],
"label": ["for", "accesskey"],
"h2": [],
"ol": [],
"ul": [],
"li": [],
"dl": [],
"dd": [],
"dt": [],
"pre": ["width"], // width only in HTML 3.2
"meta": [],
"link": ["charset", "href", "hreflang", "type", "rel", "rev", "media"],
"noscript": [],
"blockquote": [],
"div": [],
"form": ["action", "method", "enctype", "accept", "name", "accept-charset"],
"hr": ["align", "noshade", "size", "width"],
"fieldset": [],
"legend": ["accesskey"],
"address": [],
"a": ["name", "href", "accesskey", "rel", "rev", "shape", "coords", "tabindex"],
"map": ["name"],
"area": ["nohref", "shape", "coords", "href", "alt", "tabindex", "accesskey"],
"img": ["src", "alt", "longdesc", "name", "height", "width", "usemap", "ismap", "hspace", "vspace", "align"], // some attributes only in HTML 3.2
// "applet": [], // don't know what attributes this has. also, way way too crashy.
"object": ["declare", "classid", "codebase", "data", "type", "codetype", "archive", "standby", "height", "width", "usermap", "name", "tabindex"],
"embed": ["src", "type", "pluginspage"], // not in DTD??
"param": ["id", "name", "value", "valuetype", "type"],
"ins": ["cite", "datetime"],
"del": ["cite", "datetime"],

// Be nice to tables; stirtable takes care of most of this stuff.
/*
"table": ["summary", "width", "border", "frame", "rules", "cellspacing", "cellpadding"],
"caption": [],
"colgroup": ["span", "width", "align", "valign"],
"col": ["span", "width", "align", "valign"],
"thead": ["align", "valign"],
"tr": ["align", "valign"],
"td": ["align", "valign", "abbr", "axis", "headers", "scope", "rowspan", "colspan"],
"th": ["align", "valign", "abbr", "axis", "headers", "scope", "rowspan", "colspan"],
*/

"base": ["href"],
"meta": ["http-equiv", "name", "content", "scheme"],
"style": ["type", "media", "title"],
"script": ["charset", "type", "src", "defer", "event", "for"],
"html": [],
"head": [],
"body": ["bgcolor", "text", "link", "vlink", "alink", "background"], // attributes only in HTML 3.2

// evil, probably not in HTML 4 DTD for a good reason
"xmp": [], 
"listing": [],
"plaintext": [],

"isindex": [], // obsolete

// From HTML 3.2 only:
"s": [],
"basefont": ["size"],
"font": ["size", "color"],
"menu": [], // ???

// Everyone loves to hate...
"blink": [],
"marquee": ["behavior", "bgcolor", "direction", "height", "hspace", "loop", "scrollamount", "scrolldelay", "truespeed", "vspace", "width"],

// New things
"canvas": ["width", "height"],

// http://lxr.mozilla.org/seamonkey/source/editor/libeditor/html/nsHTMLEditUtils.cpp#585 -- i don't know what some of these are!
"dir": [],
"keygen": [],
"nobr": [],
"multicol": [],
"u": [],
"wbr": [],
"spacer": [],
"noframes": [],
"noembed": [],

// don't forget the lonely "unknown element"!
"foo": [],

// iframe
"frame": []
// "frameset": [], // avoid bug 324318

};



// attributes is a hash, but sometimes we want to treat it as an array-like thing.  so....
elementList = [];
for (var prop in elements)
  elementList.push(prop);

fuzzOnload();