// fuzz.js 3.0


// fuzz.js provides the following features for fuzzers that use it:
// * Seeded pseudorandom numbers
// * Chunking of steps for speed
// * Recording and playing, to make reduction possible.  (Even takes chunking into account most of the time.)
// * Bookmarklet mode
// * Things to aid automated reduction with Lithium (DD* markers, goQuitApplication)
// * Things to aid automated testing with Spider
// * Pausing (Esc pauses, Shift+Esc unpauses)
// * And, by being a "framework", consistency between fuzzers and hopefully easier maintainance.

// Rules for fuzzers:
// * Defined the following variables: fuzzerName, recordStrategy.
// * makeCommand should try to be free of side-effects.
// * doCommand should be free of randomness.
// * init should call startFuzzing (either immediately, or if needed, using timeout).
// * Use "doc", not "document" (for spider).
// * Use "win", not "window" (for spider).
//   * window.status is probably an exception.
//   * Be careful when using "if (window.foo)" to test whether variables exist.

// How to reduce a crash found with a fuzzing bookmarklet:
// 1. Save the page on which you found the crash.  (If it's an HTML page, use the "HTML only" saving mode.)
// 2. Paste the chunk of the fuzz script page containing the two <script> elements (one with src="fuzz.src", one inline) into the saved page.
//      For anything that has a document.body (HTML or XHTML), paste it in <head> (below meta charset, if present -- bug 322181).
//      For things like SVG, paste it so that it's the first child of the root, with no whitespace before it.
// 3. Load the page.
// 4. A dialog should appear asking for fuzzing parameters.  Give it the same parameters you gave the bookmarklet, and verify that it still crashes.
// 5. Do step 4 again, but add the fifth (maxSteps) parameter (overshooting a bit is ok) and make the sixth paramter 1 to tell it to record.
// 6. For some fuzzers, you'll see a big blog of pastable code.  Paste it into the testcase, replacing the lines "var fuzzSettings = null;" and "var fuzzCommmands = null;".  (For others fuzzers, you'll have to grep console output for origCount to create the fuzzCommands array; this process isn't as user-friendly.)
// 7. Verify that it still crashes.
// 8. Reduce the testcase, usually starting with the fuzzCommands array.  This is usually the most time-consuming part of a reduction, but Lithium (https://bugzilla.mozilla.org/show_bug.cgi?id=329066) can often automate it.
// 9. Reduce the rest of the testcase manually.


var dumpEachCommand = false; // Can be set to true by using the "Record as it goes" reduction strategy.  Also, some fuzzers always set this to true.
var dumpEachSeed = false; // Set to true if makeCommand has side effects, such as crashing, so you have to reduce "the hard way".
var gPageCompleted = false;

// like spidermonkey's uneval, but cross-browser, and only for strings and simple things like numbers.
function simpleSource(s)
{
  if (typeof s == "string")
    return "\"" + s.replace(/\\/g, "\\\\").replace(/\"/g, "\\\"") + "\"";
  else
    return "" + s; // hope this is right ;)  should work for numbers.
}

function fuzzOnload(ev)
{
  window.removeEventListener("load", fuzzOnload, false); // don't init twice when using a bookmarklet

  initFuzzerGeneral();

  // initFuzzerSpecific() is expected to call startFuzzing.  Most fuzzers call it immediately but some use a timeout or callback.
  initFuzzerSpecific();
}

function initFuzzerGeneral()
{
  obtainSettings();

  dumpln(fuzzerName + " running on " + win.location.href);
  dumpln("var fuzzSettings = [" + fuzzSettings + "];");

  rnd = randomizer(fuzzSeed);
}


function startFuzzing()
{
  if (window.fuzzCommands) {
    // Already recorded, so we should play.
    count = 0;
    if (recordStrategy == "Special") {
      playFunsSpecial();
    } else {
      playFunsChunk();
    }
  }
  else {
    // Not already recorded
    if (recordMode) {
      recordSomehow();
    } else {
      immedChunk(numImmediate);
    }
  }
}

function recordSomehow()
{
  switch(recordStrategy) {
    case "Record without doing": 
      recordFuns();
      break;

    case "Record as it goes":
      alert("This fuzzer records as it goes.  If you encounter a crash or other bug you want to reduce, grep console output for lines containing origCount and paste that in as a fuzzCommands array.");
      dumpEachCommand = true;
      immedChunk(numImmediate);
      break;

    case "Special":
      recordFunsSpecial(); // defined by the fuzzer
      break;

    case "None":
    default:
      alert("This fuzzer doesn't seem to support recording.");
  }
}


var immedCount = 0;

function immedChunk(changes)
{
  if (gPageCompleted) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (done!)";
  }
  else if (pauseFuzzing) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (paused)";
  }
  else {   
    // Not paused or finished.
    
    if (immedCount > maxSteps) {
      gPageCompleted = true; // Tell Spider we're done.
      pauseFuzzing = true; // Don't keep printing useless stuff every 100 steps, don't keep trying to load the next page in the scan.
    } 
    else {

      for (var i = 0; i < changes; ++i) {
        ++immedCount;
        if (!pauseFuzzing && immedCount <= maxSteps)
          immedStep();
      }

      if (dumpEachCommand) {
        dumpln("  { origCount: " + immedCount + ", rest: true },");
      }

    }
  }

  // gPageCompleted may have been set before immedChunk() was called, or
  // as a result of immedStep(), or by code *in* immedChunk!?
  if (gPageCompleted) {
     if (getURLParam("scan")) {
      // Take advantage of seed being the first parameter to replace it easily.
      location.search = location.search.replace("fuzz=" + fuzzSeed, "fuzz=" + (fuzzSeed + 1));
    }
  }
  else if (stepsPerInterval != 0)
    setTimeout(function() { immedChunk(stepsPerInterval); }, interval);
}





function immedStep()
{
  if (dumpEachSeed) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { rnd.setSeed(" + rnd.getSeed() + "); immedStep(); } },");
  }
  
  var x = makeCommand();
  window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount;
  // randomJS.collectGarbage(); // Whee!  I put this right after "window.status" because that's what I see in the crashes...

  if (dumpEachCommand) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { " + x + " } },");
  } else {
    // Dump stuff occasionally so that if we hit a crash, we have a hint as to where we hit it.
    if (immedCount % (100) == 0)
      dumpln(fuzzerName + ": " + immedCount);
  }


  try {
    doCommand(x);
    // how will eval work? will "this" be the fuzzer?
  }
  catch(e) { 
    // Should this report the error somehow?
  }
}



/************
 * SETTINGS *
 ************/
 
var fuzzSeed = 0;
var numImmediate = 0;
var stepsPerInterval = 100;
var interval = 400;
var maxSteps = Infinity;
var recordMode = 0;


function obtainSettings()
{

  // if (window.fuzzSettings === undefined) alert("No settings variable declaration? No fuzzer?");
  // Let's just allow this to hit an error, instead of putting up an annoying alert whenever a fuzzer has a typo or we're reducing like crazy but keeping goQuitApplication.
  /*
  if (fuzzSettings == null)
    fuzzSettings = getURLParam("fuzz");

  if (fuzzSettings == null)
    fuzzSettings = prompt(fuzzerName + " settings\n\nSeed\nNumber of changes in first chunk\nNumber of changes in each subsequent chunk\nInterval between chunks in milliseconds\nTotal number of changes (or 0 for unlimited)\n1 to record, 0 to go", "0, 0, 100, 400, 0, 0");

  if (fuzzSettings == null) {
    fuzzSeed = 0; // for playing around with javascript: URLs, accessing allNodes, etc.
    numImmediate = stepsPerInterval = maxSteps = 0;
    interval = 1000;
    // throw "User hit cancel!";
    return;
  }

  if (typeof fuzzSettings == "string") {
    fuzzSettings = fuzzSettings.replace(/;/g, ","); // allow semicolons in place of commas (makes Talkback's URL field happier)
    fuzzSettings = eval("[" + fuzzSettings + "]");
  }

  if (fuzzSettings[0]) fuzzSeed         = fuzzSettings[0];
  if (fuzzSettings[1]) numImmediate     = fuzzSettings[1];
  if (fuzzSettings[2]) stepsPerInterval = fuzzSettings[2];
  if (fuzzSettings[3]) interval         = fuzzSettings[3];
  if (fuzzSettings[4]) maxSteps         = fuzzSettings[4]; // if it's 0, it remains Infinity!
  if (fuzzSettings[5]) recordMode       = fuzzSettings[5];
  // "force quit" option should be here, not in the pasted fuzzCommands array!
//0, 0, 100, 400, 0, 0
*/


 fuzzSeed         = 0;
numImmediate     = 0;
 stepsPerInterval = 100;
 interval         = 400;
 maxSteps         = 0; // if it's 0, it remains Infinity!
  recordMode       = 0;
}

function getURLParam(p)
{
  var s = win.location.search + "&"; 
  var r = new RegExp("[?&]" + p + "\\=([^&]*)\\&");
  if(r.exec(s) == null)
    return null;
  var raw = RegExp.$1;
  return unescape(raw.replace(/\+/g," "));
}


/***********
 * PAUSING *
 ***********/
 
var pauseFuzzing = false;

function escPause(ev)
{
  if (ev.keyCode == 27) { // esc
    if (!ev.shiftKey) {
      dumpln("Pausing fuzzing.");
      pauseFuzzing = true;
    } else {
      // Shift+Esc (works in Firefox but not in Safari)
      dumpln("Resuming fuzzing.");
      pauseFuzzing = false;
    }
  }
}

window.addEventListener("keypress", escPause, false);




/*********************************
 * SIMPLE RECORDING AND PLAYBACK *
 *********************************/
 
// This section is for when recordStrategy == "Record without doing".

function playFuns()
{
  var command = fuzzCommands[count];

  ++count;

  if (command == undefined)
    return false;
  
  var message = fuzzerName + " (playing): origCount " + command.origCount;
  window.status = message;

  // Also dump sometimes, in case Firefox has stopped painting, or in case we want to see info related to a crash *before reducing*
  if (command.origCount % (100) == 0 && !command.rest)
    dumpln(fuzzerName + " (playing): origCount " + command.origCount);

  if (command.fun) {
    try { 
      command.fun(); 
    } catch(e) {
      var errorAsString = "";
      try {
        errorAsString = "" + e;
      } catch(e2) { 
        errorAsString = "[cannot convert this exception to string!]";
      }
      window.status += " threw: " + errorAsString; 
    }
  }

  return !(command.rest);
}



function playFunsChunk()
{
  while(playFuns())
    ;

  // Set a timer for the next chunk
  setTimeout(playFunsChunk, interval);
}


function recordFuns()
{
  // This is nice in that it avoids verbose output...
  var fuzzCommands = [];

  var output =
     "var fuzzSettings = [" + fuzzSettings + "];\n" +
     "var fuzzCommands = [\n" +
     "// DD" + "BEGIN\n";
     
  for (var count = 1; count <= maxSteps; ++count)
  {
    var command = makeCommand();
    output += "  { origCount: " + count + ", fun: function() { " + command + " } },\n";

    var countAfterImmed = count - numImmediate;
    if((countAfterImmed >= 0) && (countAfterImmed % stepsPerInterval == 0)) {
      output += "  { origCount: " + count + ", rest: true },\n";
    }
  }
  
  output += 
    "// DD" + "END\n" +
    "];";

  giveToCopy("// Paste this into the script, replacing the first two lines:\n\n" + output);
  
  // Verbosity of output is due to in part to
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322175 and
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322177

}




/***************************
 * REPRODUCIBLE RANDOMNESS *
 ***************************/
 

// This random number generator obviously repeats pretty often :(  Are there better RNGs available in (or for) JavaScript?

// Given a seed, returns a function that generates random integers.
// Based on the Central Randomizer 1.3 (C) 1997 by Paul Houle (paul@honeylocust.com)
// See: http://www.honeylocust.com/javascript/randomizer.html
function randomizer(seed)
{
  
  // Returns a somewhat random real number between 0 (inclusive) and 1 (not inclusive), like Math.random() but with a seed I control.
  // Modulus is in parens to work around a Crunchinator bug.
  function rndReal() {
	seed = (seed*9301 + 49297) % (233280);
	return seed/(233280.0);
  };

  // Returns a somewhat random integer in the range {0, 1, 2, ..., number-1}
  function rndInt(number) {
	return Math.floor(rndReal()*number);
  };
  
  // for hard reductions only
  rndInt.setSeed = function(s) { seed = s; };
  rndInt.getSeed = function() { return seed; };
  
  return rndInt;
}


// Based on http://archives.hwg.org/hwg-basics/002d01c11d20$c3dec4c0$0200000a@matthew
// Nope, this is even worse (e.g. with seed = 4)
/*
function randomizer(seed)
{
  function rndInt(number)
  {
    seed = (0x015a4e35 * seed) % (0x7fffffff);
    return (seed >> 16) % number;
  }
  
  return rndInt;
}
*/


function rndElt(a) { return a[rnd(a.length)]; }



/**********
 * OUTPUT *
 **********/
 
function dumpln (s)
{
  if (window.dump) 
    dump(s + "\n");
}


function giveToCopy(s)
{
  // This can crash on Linux (due to a long string in the address bar?), so we should have an alternate strategy...
  location.href = "data:text/plain;charset=UTF-8," + encodeURIComponent(s);
  
  // this is nice, but only works is HTML (see bug 332848): 
  //document.open("text/plain");
  //document.write(s);
  
  // other strategies: add an <html:pre> to the end of the document, ...
  
  // When anything else would be too painful
  //dumpln(s);
  //alert("Dumped");
}




/**************
 * FORCE QUIT *
 **************/

// From quit.js, which Bob Clary extracted from mozilla/toolkit/content

function canQuitApplication()
{
  var os = Components.classes["@mozilla.org/observer-service;1"]
    .getService(Components.interfaces.nsIObserverService);
  if (!os) 
  {
    return true;
  }

  try 
 {
    var cancelQuit = Components.classes["@mozilla.org/supports-PRBool;1"]
      .createInstance(Components.interfaces.nsISupportsPRBool);
    os.notifyObservers(cancelQuit, "quit-application-requested", null);
  
    // Something aborted the quit process. 
    if (cancelQuit.data)
    {
      return false;
    }
  }
  catch (ex) 
  {
  }
  os.notifyObservers(null, "quit-application-granted", null);
  return true;
}

function goQuitApplication()
{
  var privs = 'UniversalPreferencesRead UniversalPreferencesWrite ' +
    'UniversalXPConnect';

  try
  {
    netscape.security.PrivilegeManager.enablePrivilege(privs);
  }
  catch(ex)
  {
    throw('goQuitApplication: privilege failure ' + ex);
  }

  if (!canQuitApplication())
  {
    return false;
  }

  var kAppStartup = '@mozilla.org/toolkit/app-startup;1';
  var kAppShell   = '@mozilla.org/appshell/appShellService;1';
  var   appService;
  var   forceQuit;

  if (kAppStartup in Components.classes)
  {
    appService = Components.classes[kAppStartup].
      getService(Components.interfaces.nsIAppStartup);
    forceQuit  = Components.interfaces.nsIAppStartup.eForceQuit;

  }
  else if (kAppShell in Components.classes)
  {
    appService = Components.classes[kAppShell].
      getService(Components.interfaces.nsIAppShellService);
    forceQuit = Components.interfaces.nsIAppShellService.eForceQuit;
  }
  else
  {
    throw 'goQuitApplication: no AppStartup/appShell';
  }

  var windowManager = Components.
    classes['@mozilla.org/appshell/window-mediator;1'].getService();

  var windowManagerInterface = windowManager.
    QueryInterface(Components.interfaces.nsIWindowMediator);

  var enumerator = windowManagerInterface.getEnumerator(null);

  while (enumerator.hasMoreElements())
  {
    var domWindow = enumerator.getNext();
    if (("tryToClose" in domWindow) && !domWindow.tryToClose())
    {
      return false;
    }
    domWindow.close();
  }

  try
  {
    appService.quit(forceQuit);
  }
  catch(ex)
  {
    throw('goQuitApplication: ' + ex);
  }

  return true;
}



/***************
 * OTHER MODES *  
 ***************/


var win = (window.gSpider ? gSpider.mDocument.defaultView : window); // For spider.  I hope this is right (and safe).
var doc = win.document;


// Non-bookmarklet
//window.addEventListener("load", fuzzOnload, false); // !!!!

// Bookmarklet
// fuzzOnload();



/********************
 * COMMON DOM STUFF *
 ********************/
 
function addDOMNodes(n, enterIFrames, enterSVG, enterXBL) { 
  var i,c;

  if (n.tagName && n.tagName.toLowerCase().indexOf("script" != -1) && n.getAttribute("id") && n.getAttribute("id").indexOf("fuzz") != -1) {
    if (n.previousSibling && n.previousSibling.nodeType == Node.TEXT_NODE)
      alert("Warning: fuzz script found with a text (whitespace?) node before it.  This could affect node counts.");
    return;
  }

  if (n.tagName && n.tagName.toLowerCase() == "frameset") {
    gPageCompleted = true;
    throw "Avoiding bug 324318!";
  }
  

  // Add this node
  allNodes.push(n); 
  
  // Look for children
  for (i=0;c=n.childNodes[i];++i) {
    //try {
      addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
    //} catch(e) { /* Why would this fail? */ }
  }
  
  // If this is a frame, look inside it.
  if(enterIFrames) {
    try {
      if(n.contentDocument)
        try {
          addDOMNodes(n.contentDocument.documentElement, enterIFrames, enterSVG, enterXBL); 
        } catch(ex) { /* third-party iframe */ }
    } catch(ex) { /* avoid bug 310994 */ }
  }
  
  // If this is an <embed> referencing an SVG image, look inside it for SVG nodes.
  if (enterSVG) {
    if (n.getSVGDocument) { 
      try{
        if (n.getSVGDocument())
          addDOMNodes(n.getSVGDocument(), enterIFrames, enterSVG, enterXBL);
      } catch(ex) { }
    }
  }

  // Look for anonymous children (XBL, at least)
  // try/catch in case browser doesn't support getAnonymousNodes, or in case Firefox decides to throw a FAILURE error (!?).
  if (enterXBL) {
    try { 
      if (n.nodeType==1 && doc.getAnonymousNodes(n)) {
        for(i=0;c=doc.getAnonymousNodes(n)[i];++i) {
          if (!this.anonymousCount)
            anonymousCount = 0;
          ++anonymousCount;
          addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
        }
      }
    } catch (ex) { } 
  }
}

// === Paste output from first run here, replacing the following two lines: ===

var fuzzSettings = null;
var fuzzCommands = null;

// === End pasteover ===


var fuzzerName = "Random Classes 3.0";
var recordStrategy = "Record without doing";
var suggestedMaxSteps = 2000; // very active fuzzer!

// Based on the CSSGen list, which was in turn based on the Random Styles list.
var CSSProperties = [

  {
    prop: "float",
    values: ["left", "right", "none"]
  },
  {
    prop: "clear",
    values: ["none", "left", "right", "both"]
  },
  {
    prop: "display",
    values: ["inline", "block", "-moz-inline-block", "list-item",
    // Table
    "table", "table-row-group", "table-header-group", "table-footer-group", "table-row", "table-column-group",
    "table-column", "table-cell" /* , "table-caption" -- avoid hitting bug 348126 */
    // XUL
    /* , "-moz-box", "-moz-inline-grid", "-moz-grid", "-moz-box", "-moz-inline-box",
    "-moz-grid-group", "-moz-grid-line", "-moz-stack", "-moz-inline-stack", "-moz-deck", "-moz-popup", "-moz-groupbox" */
    ]
  },
  {
    prop: "visibility",
    values: ["visible", "hidden", "collapse"]
  },
  {
    prop: "position",
    values: ["static", "relative", "absolute", "fixed"]
  },
  {
    prop: "overflow",
    values: ["visible", "scroll", "hidden", "auto", "-moz-scrollbars-horizontal", "-moz-scrollbars-none", "-moz-scrollbars-vertical", "-moz-hidden-unscrollable"]
  },
  {
    prop: "color",
    values: ["red", "green", "blue", "black"]
  },
  {
    prop: "background",
    values: ["transparent", "#fcd", "url(http://www.google.com/images/logo_sm.gif)", "url(https://bugzilla.mozilla.org/attachment.cgi?id=165273&action=view)"]
  },
  {
    prop: "width",
    values: ["20em", "1px", "auto", "10%", "200%"]
  },
  {
    prop: "height",
    values: ["2em", "100px", "auto", "10%", "200%"]
  },
  {
    prop: "content",
    values: [
    
     "'generated'", 
     "'generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated generated'", 
     "url(http://www.mozilla.org/images/mozilla-16.png)", 
     "url(https://bugzilla.mozilla.org/attachment.cgi?id=165273&action=view)",


// Multiline generated content triggers some special code in Gecko, so let's lets it well!

// inspired by uri's attachment in bug 343540. the amount of escaping here is ludicrous; one level of backslashes is for this being a string and the other is for it being eval'ed as a string later.

// why so many zeroes? "If a character in the range [0-9a-fA-F] follows the hexadecimal number, the end of the number needs to be made clear. There are two ways to do that... (2) by providing exactly 6 hexadecimal digits."

     "'XXX-\\\\00000aYYY'", 
     "'XXX\\\\00000aYYY'",
     "'A\\\\00000aB\\\\00000aC\\\\00000aD\\\\00000aE'",
     "'a \\\\00000ab \\\\00000ac \\\\00000ad \\\\00000ae'"
    ]
  },
  {
    prop: "padding",
    values: ['0','10px','2em','100px','3pt','auto']
  },
  {
    prop: "margin",
    values: ['0','10px','2em','100px','3pt','auto']
  },
  {
    prop: "opacity",
    values: ["0.2", "1"]
  },
  {
    prop: "letter-spacing",
    values: ["4em", "inherit", "1px", "-2px", "-1em", "-2em"]
  },
  {
    prop: "z-index",
    values: ["inherit", "5", "0", "-200"]
  },
/*
  // Commented out to avoid hitting bug 334147
  {
    prop: "direction",
    values: ["inherit", "ltr", "rtl"]
  },
*/
  {
    prop: "text-align",
    values: ["left", "right", "justify"]
  },
  {
    prop: "border-collapse",
    values: ["inherit", "collapse"]
  },
  {
    prop: "outline",
    values: ["1px solid yellow", "none"]
  },
  {
    prop: "font-size",
    values: ["130%", "80%"]
  },
  
// Column stuff is commented out due to frequent crashes, perhaps including bug 337419.
/*
  {
    prop: "-moz-column-count",
    values: ["-1", "0", "1", "2", "3", "15"]
  },
  {
    prop: "-moz-column-width",
    values: ["20em", "1px", "auto", "10%", "200%"]
  },
  {
    prop: "-moz-column-gap",
    values: ["20em", "1px", "auto", "10%", "200%"]
  },
*/

  { prop: "top", values: ["20em", "-20em", "0", "10%", "5px", "-5px"] },
  { prop: "bottom", values: ["20em", "-20em", "0", "10%", "5px", "-5px"] },
  { prop: "right", values: ["20em", "-20em", "0", "10%", "5px", "-5px"] },
  { prop: "left", values: ["20em", "-20em", "0", "10%", "5px", "-5px"] },

  {
    prop: "white-space",
    values: ["inherit", "normal", "pre", "-moz-pre-wrap" /* also -o-pre-wrap and -khtml-pre-wrap and pre-wrap */, "nowrap"]
  } // ,

];

var pseudoElements = [ ":after", ":before", ":first-letter", ":first-line", "::-moz-selection", "::-moz-focus-inner", "::-moz-focus-outer", "::-moz-list-bullet", "::-moz-list-number", "::-moz-hframeset-border", "::-moz-vframeset-border",

/* The following are from nsCSSAnonBoxList.h in Mozilla source code.  ::-moz-column-content is known to work for color and visibility! */
// Commented out because I think bug 331883 took care of these.
// "::-moz-non-element", "::-moz-anonymous-block", "::-moz-anonymous-positioned-block", "::-moz-first-line-fixup", "::-moz-line-frame", "::-moz-buttonlabel", "::-moz-cell-content", "::-moz-dropdown-list", "::-moz-fieldset-content", "::-moz-frameset-blank", "::-moz-radio", "::-moz-checkbox", "::-moz-display-comboboxcontrol-frame", "::-moz-table", "::-moz-table-cell", "::-moz-table-column-group", "::-moz-table-column", "::-moz-table-outer", "::-moz-table-row-group", "::-moz-table-row", "::-moz-canvas", "::-moz-pagebreak", "::-moz-page", "::-moz-pagecontent", "::-moz-page-sequence", "::-moz-scrolled-content", "::-moz-scrolled-canvas", "::-moz-scrolled-page-sequence", "::-moz-column-content", "::-moz-viewport", "::-moz-viewport-scroll", "::-moz-select-scrolled-content", "::-moz-tree-column", "::-moz-tree-row", "::-moz-tree-separator", "::-moz-tree-cell", "::-moz-tree-indentation", "::-moz-tree-line", "::-moz-tree-twisty", "::-moz-tree-image", "::-moz-tree-cell-text", "::-moz-tree-checkbox", "::-moz-tree-progressmeter", "::-moz-tree-drop-feedback", "::-moz-math-stretchy", "::-moz-math-anonymous", "::-moz-math-inline"
 ];


// I don't think the strategy of using CSS on an attribute called "class" works for MathML :(
// Could use IDs in the same way I currently use classes, but that seems hacky and messy.
var classes = [ "cat", "toad", "zebra", "lizard", "penguin", "elephant" ];


function randomSelector() {
  var selector;
  
  if (rnd(30) == 0)
    selector = "*";
  else
    selector = "." + rndElt(classes);

  if (rnd(2) == 0)
    selector += rndElt(pseudoElements);

  return selector;
}

function randomRule() {
  var rule = randomSelector() + " { ";
  var i, pair;
  for (i = 0; pair = CSSProperties[i]; ++i)
    if (rnd(6) == 0)
      rule += pair.prop + ": " + rndElt(pair.values) + "; ";
  rule += "}";
  return rule;
}

function randomStylesheet()
{
  // Just one rule per stylesheet for now.
  var s = "";
  for (var i = 0; i < 1; ++i)
    s += randomRule();
  return s;
}

function makeCommand()
{
  switch(rnd(2)) {
    case 0:
      // Change one of several stylesheets.
      return "setTextContent(stylesheets[" + rnd(stylesheets.length) + "], " + simpleSource(randomStylesheet()) + ");";
    default:
      // Change a random node's class.
      // return "allElements[" + rnd(allElements.length) + "].className = " + simpleSource(rndElt(classes)) + ";";
      return "allElements[" + rnd(allElements.length) + "].setAttribute('class', " + simpleSource(rndElt(classes)) + ");";
  }
}

function doCommand(s) { eval(s); }

function setTextContent(n, s)
{
  // Safari sucks, so we can't use foo.textContent.
  while (n.firstChild)
    n.removeChild(n.firstChild);
    
  addTextContent(n, s);
}

function addTextContent(n, s)
{
  n.appendChild(doc.createTextNode(s));
}

var stylesheets = [];

var allElements = [];

function addElements(n) {

  if (n.tagName && n.tagName.toLowerCase().indexOf("script" != -1) && n.getAttribute("id") && n.getAttribute("id").indexOf("fuzz") != -1) {
    if (n.previousSibling && n.previousSibling.nodeType == Node.TEXT_NODE)
      alert("Warning: fuzz script found with a whitespace node before it.  This could affect node counts.");
    return;
  }

  
  // MathML does not expose style, also doesn't use class attribute the way i want
  /*
  try { 
    if (n.style)
      allElements.push(n);
  } catch(e) { }
  */
  
  allElements.push(n);

  var i,c;
  for (i=0;c=n.childNodes[i];++i)
    if (c.nodeType == Node.ELEMENT_NODE)
      addElements(c);
};

function initFuzzerSpecific()
{
  var myRoot = doc.getElementsByTagName("body")[0] || doc.documentElement; 
  addElements(myRoot);

  var myStylesheetHolder = doc.getElementsByTagName("head")[0] || myRoot;

  for (var i = 0; i < 25; ++i) {
    var s = doc.createElementNS("http://www.w3.org/1999/xhtml", 'style');
    s.style.display = "none";
    myStylesheetHolder.appendChild(s);
    stylesheets.push(s);
  }
  
  startFuzzing();
}

fuzzOnload();