// fuzz.js 3.0


// fuzz.js provides the following features for fuzzers that use it:
// * Seeded pseudorandom numbers
// * Chunking of steps for speed
// * Recording and playing, to make reduction possible.  (Even takes chunking into account most of the time.)
// * Bookmarklet mode
// * Things to aid automated reduction with Lithium (DD* markers, goQuitApplication)
// * Things to aid automated testing with Spider
// * Pausing (Esc pauses, Shift+Esc unpauses)
// * And, by being a "framework", consistency between fuzzers and hopefully easier maintainance.

// Rules for fuzzers:
// * Defined the following variables: fuzzerName, recordStrategy.
// * makeCommand should try to be free of side-effects.
// * doCommand should be free of randomness.
// * init should call startFuzzing (either immediately, or if needed, using timeout).
// * Use "doc", not "document" (for spider).
// * Use "win", not "window" (for spider).
//   * window.status is probably an exception.
//   * Be careful when using "if (window.foo)" to test whether variables exist.

// How to reduce a crash found with a fuzzing bookmarklet:
// 1. Save the page on which you found the crash.  (If it's an HTML page, use the "HTML only" saving mode.)
// 2. Paste the chunk of the fuzz script page containing the two <script> elements (one with src="fuzz.src", one inline) into the saved page.
//      For anything that has a document.body (HTML or XHTML), paste it in <head> (below meta charset, if present -- bug 322181).
//      For things like SVG, paste it so that it's the first child of the root, with no whitespace before it.
// 3. Load the page.
// 4. A dialog should appear asking for fuzzing parameters.  Give it the same parameters you gave the bookmarklet, and verify that it still crashes.
// 5. Do step 4 again, but add the fifth (maxSteps) parameter (overshooting a bit is ok) and make the sixth paramter 1 to tell it to record.
// 6. For some fuzzers, you'll see a big blog of pastable code.  Paste it into the testcase, replacing the lines "var fuzzSettings = null;" and "var fuzzCommmands = null;".  (For others fuzzers, you'll have to grep console output for origCount to create the fuzzCommands array; this process isn't as user-friendly.)
// 7. Verify that it still crashes.
// 8. Reduce the testcase, usually starting with the fuzzCommands array.  This is usually the most time-consuming part of a reduction, but Lithium (https://bugzilla.mozilla.org/show_bug.cgi?id=329066) can often automate it.
// 9. Reduce the rest of the testcase manually.


var dumpEachCommand = false; // Can be set to true by using the "Record as it goes" reduction strategy.  Also, some fuzzers always set this to true.
var dumpEachSeed = false; // Set to true if makeCommand has side effects, such as crashing, so you have to reduce "the hard way".
var gPageCompleted = false;

// like spidermonkey's uneval, but cross-browser, and only for strings and simple things like numbers.
function simpleSource(s)
{
  if (typeof s == "string")
    return "\"" + s.replace(/\\/g, "\\\\").replace(/\"/g, "\\\"") + "\"";
  else
    return "" + s; // hope this is right ;)  should work for numbers.
}

function fuzzOnload(ev)
{
  window.removeEventListener("load", fuzzOnload, false); // don't init twice when using a bookmarklet

  initFuzzerGeneral();

  // initFuzzerSpecific() is expected to call startFuzzing.  Most fuzzers call it immediately but some use a timeout or callback.
  initFuzzerSpecific();
}

function initFuzzerGeneral()
{
  obtainSettings();

  dumpln(fuzzerName + " running on " + win.location.href);
  dumpln("var fuzzSettings = [" + fuzzSettings + "];");

  rnd = randomizer(fuzzSeed);
}


function startFuzzing()
{
  if (window.fuzzCommands) {
    // Already recorded, so we should play.
    count = 0;
    if (recordStrategy == "Special") {
      playFunsSpecial();
    } else {
      playFunsChunk();
    }
  }
  else {
    // Not already recorded
    if (recordMode) {
      recordSomehow();
    } else {
      immedChunk(numImmediate);
    }
  }
}

function recordSomehow()
{
  switch(recordStrategy) {
    case "Record without doing": 
      recordFuns();
      break;

    case "Record as it goes":
      alert("This fuzzer records as it goes.  If you encounter a crash or other bug you want to reduce, grep console output for lines containing origCount and paste that in as a fuzzCommands array.");
      dumpEachCommand = true;
      immedChunk(numImmediate);
      break;

    case "Special":
      recordFunsSpecial(); // defined by the fuzzer
      break;

    case "None":
    default:
      alert("This fuzzer doesn't seem to support recording.");
  }
}


var immedCount = 0;

function immedChunk(changes)
{
  if (gPageCompleted) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (done!)";
  }
  else if (pauseFuzzing) {
    window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount + " (paused)";
  }
  else {   
    // Not paused or finished.
    
    if (immedCount > maxSteps) {
      gPageCompleted = true; // Tell Spider we're done.
      pauseFuzzing = true; // Don't keep printing useless stuff every 100 steps, don't keep trying to load the next page in the scan.
    } 
    else {

      for (var i = 0; i < changes; ++i) {
        ++immedCount;
        if (!pauseFuzzing && immedCount <= maxSteps)
          immedStep();
      }

      if (dumpEachCommand) {
        dumpln("  { origCount: " + immedCount + ", rest: true },");
      }

    }
  }

  // gPageCompleted may have been set before immedChunk() was called, or
  // as a result of immedStep(), or by code *in* immedChunk!?
  if (gPageCompleted) {
     if (getURLParam("scan")) {
      // Take advantage of seed being the first parameter to replace it easily.
      location.search = location.search.replace("fuzz=" + fuzzSeed, "fuzz=" + (fuzzSeed + 1));
    }
  }
  else if (stepsPerInterval != 0)
    setTimeout(function() { immedChunk(stepsPerInterval); }, interval);
}





function immedStep()
{
  if (dumpEachSeed) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { rnd.setSeed(" + rnd.getSeed() + "); immedStep(); } },");
  }
  
  var x = makeCommand();
  window.status = fuzzerName + " (" + fuzzSettings + ") : " + immedCount;
  // randomJS.collectGarbage(); // Whee!  I put this right after "window.status" because that's what I see in the crashes...

  if (dumpEachCommand) {
    dumpln("  { origCount: " + immedCount + ", fun: function() { " + x + " } },");
  } else {
    // Dump stuff occasionally so that if we hit a crash, we have a hint as to where we hit it.
    if (immedCount % (100) == 0)
      dumpln(fuzzerName + ": " + immedCount);
  }


  try {
    doCommand(x);
    // how will eval work? will "this" be the fuzzer?
  }
  catch(e) { 
    // Should this report the error somehow?
  }
}



/************
 * SETTINGS *
 ************/
 
var fuzzSeed = 0;
var numImmediate = 0;
var stepsPerInterval = 100;
var interval = 400;
var maxSteps = Infinity;
var recordMode = 0;


function obtainSettings()
{

  // if (window.fuzzSettings === undefined) alert("No settings variable declaration? No fuzzer?");
  // Let's just allow this to hit an error, instead of putting up an annoying alert whenever a fuzzer has a typo or we're reducing like crazy but keeping goQuitApplication.
  /*
  if (fuzzSettings == null)
    fuzzSettings = getURLParam("fuzz");

  if (fuzzSettings == null)
    fuzzSettings = prompt(fuzzerName + " settings\n\nSeed\nNumber of changes in first chunk\nNumber of changes in each subsequent chunk\nInterval between chunks in milliseconds\nTotal number of changes (or 0 for unlimited)\n1 to record, 0 to go", "0, 0, 100, 400, 0, 0");

  if (fuzzSettings == null) {
    fuzzSeed = 0; // for playing around with javascript: URLs, accessing allNodes, etc.
    numImmediate = stepsPerInterval = maxSteps = 0;
    interval = 1000;
    // throw "User hit cancel!";
    return;
  }

  if (typeof fuzzSettings == "string") {
    fuzzSettings = fuzzSettings.replace(/;/g, ","); // allow semicolons in place of commas (makes Talkback's URL field happier)
    fuzzSettings = eval("[" + fuzzSettings + "]");
  }

  if (fuzzSettings[0]) fuzzSeed         = fuzzSettings[0];
  if (fuzzSettings[1]) numImmediate     = fuzzSettings[1];
  if (fuzzSettings[2]) stepsPerInterval = fuzzSettings[2];
  if (fuzzSettings[3]) interval         = fuzzSettings[3];
  if (fuzzSettings[4]) maxSteps         = fuzzSettings[4]; // if it's 0, it remains Infinity!
  if (fuzzSettings[5]) recordMode       = fuzzSettings[5];
  // "force quit" option should be here, not in the pasted fuzzCommands array!
//0, 0, 100, 400, 0, 0
*/


 fuzzSeed         = 0;
numImmediate     = 0;
 stepsPerInterval = 100;
 interval         = 400;
 maxSteps         = 0; // if it's 0, it remains Infinity!
  recordMode       = 0;
}

function getURLParam(p)
{
  var s = win.location.search + "&"; 
  var r = new RegExp("[?&]" + p + "\\=([^&]*)\\&");
  if(r.exec(s) == null)
    return null;
  var raw = RegExp.$1;
  return unescape(raw.replace(/\+/g," "));
}


/***********
 * PAUSING *
 ***********/
 
var pauseFuzzing = false;

function escPause(ev)
{
  if (ev.keyCode == 27) { // esc
    if (!ev.shiftKey) {
      dumpln("Pausing fuzzing.");
      pauseFuzzing = true;
    } else {
      // Shift+Esc (works in Firefox but not in Safari)
      dumpln("Resuming fuzzing.");
      pauseFuzzing = false;
    }
  }
}

window.addEventListener("keypress", escPause, false);




/*********************************
 * SIMPLE RECORDING AND PLAYBACK *
 *********************************/
 
// This section is for when recordStrategy == "Record without doing".

function playFuns()
{
  var command = fuzzCommands[count];

  ++count;

  if (command == undefined)
    return false;
  
  var message = fuzzerName + " (playing): origCount " + command.origCount;
  window.status = message;

  // Also dump sometimes, in case Firefox has stopped painting, or in case we want to see info related to a crash *before reducing*
  if (command.origCount % (100) == 0 && !command.rest)
    dumpln(fuzzerName + " (playing): origCount " + command.origCount);

  if (command.fun) {
    try { 
      command.fun(); 
    } catch(e) {
      var errorAsString = "";
      try {
        errorAsString = "" + e;
      } catch(e2) { 
        errorAsString = "[cannot convert this exception to string!]";
      }
      window.status += " threw: " + errorAsString; 
    }
  }

  return !(command.rest);
}



function playFunsChunk()
{
  while(playFuns())
    ;

  // Set a timer for the next chunk
  setTimeout(playFunsChunk, interval);
}


function recordFuns()
{
  // This is nice in that it avoids verbose output...
  var fuzzCommands = [];

  var output =
     "var fuzzSettings = [" + fuzzSettings + "];\n" +
     "var fuzzCommands = [\n" +
     "// DD" + "BEGIN\n";
     
  for (var count = 1; count <= maxSteps; ++count)
  {
    var command = makeCommand();
    output += "  { origCount: " + count + ", fun: function() { " + command + " } },\n";

    var countAfterImmed = count - numImmediate;
    if((countAfterImmed >= 0) && (countAfterImmed % stepsPerInterval == 0)) {
      output += "  { origCount: " + count + ", rest: true },\n";
    }
  }
  
  output += 
    "// DD" + "END\n" +
    "];";

  giveToCopy("// Paste this into the script, replacing the first two lines:\n\n" + output);
  
  // Verbosity of output is due to in part to
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322175 and
  // https://bugzilla.mozilla.org/show_bug.cgi?id=322177

}




/***************************
 * REPRODUCIBLE RANDOMNESS *
 ***************************/
 

// This random number generator obviously repeats pretty often :(  Are there better RNGs available in (or for) JavaScript?

// Given a seed, returns a function that generates random integers.
// Based on the Central Randomizer 1.3 (C) 1997 by Paul Houle (paul@honeylocust.com)
// See: http://www.honeylocust.com/javascript/randomizer.html
function randomizer(seed)
{
  
  // Returns a somewhat random real number between 0 (inclusive) and 1 (not inclusive), like Math.random() but with a seed I control.
  // Modulus is in parens to work around a Crunchinator bug.
  function rndReal() {
	seed = (seed*9301 + 49297) % (233280);
	return seed/(233280.0);
  };

  // Returns a somewhat random integer in the range {0, 1, 2, ..., number-1}
  function rndInt(number) {
	return Math.floor(rndReal()*number);
  };
  
  // for hard reductions only
  rndInt.setSeed = function(s) { seed = s; };
  rndInt.getSeed = function() { return seed; };
  
  return rndInt;
}


// Based on http://archives.hwg.org/hwg-basics/002d01c11d20$c3dec4c0$0200000a@matthew
// Nope, this is even worse (e.g. with seed = 4)
/*
function randomizer(seed)
{
  function rndInt(number)
  {
    seed = (0x015a4e35 * seed) % (0x7fffffff);
    return (seed >> 16) % number;
  }
  
  return rndInt;
}
*/


function rndElt(a) { return a[rnd(a.length)]; }



/**********
 * OUTPUT *
 **********/
 
function dumpln (s)
{
  if (window.dump) 
    dump(s + "\n");
}


function giveToCopy(s)
{
  // This can crash on Linux (due to a long string in the address bar?), so we should have an alternate strategy...
  location.href = "data:text/plain;charset=UTF-8," + encodeURIComponent(s);
  
  // this is nice, but only works is HTML (see bug 332848): 
  //document.open("text/plain");
  //document.write(s);
  
  // other strategies: add an <html:pre> to the end of the document, ...
  
  // When anything else would be too painful
  //dumpln(s);
  //alert("Dumped");
}




/**************
 * FORCE QUIT *
 **************/

// From quit.js, which Bob Clary extracted from mozilla/toolkit/content

function canQuitApplication()
{
  var os = Components.classes["@mozilla.org/observer-service;1"]
    .getService(Components.interfaces.nsIObserverService);
  if (!os) 
  {
    return true;
  }

  try 
 {
    var cancelQuit = Components.classes["@mozilla.org/supports-PRBool;1"]
      .createInstance(Components.interfaces.nsISupportsPRBool);
    os.notifyObservers(cancelQuit, "quit-application-requested", null);
  
    // Something aborted the quit process. 
    if (cancelQuit.data)
    {
      return false;
    }
  }
  catch (ex) 
  {
  }
  os.notifyObservers(null, "quit-application-granted", null);
  return true;
}

function goQuitApplication()
{
  var privs = 'UniversalPreferencesRead UniversalPreferencesWrite ' +
    'UniversalXPConnect';

  try
  {
    netscape.security.PrivilegeManager.enablePrivilege(privs);
  }
  catch(ex)
  {
    throw('goQuitApplication: privilege failure ' + ex);
  }

  if (!canQuitApplication())
  {
    return false;
  }

  var kAppStartup = '@mozilla.org/toolkit/app-startup;1';
  var kAppShell   = '@mozilla.org/appshell/appShellService;1';
  var   appService;
  var   forceQuit;

  if (kAppStartup in Components.classes)
  {
    appService = Components.classes[kAppStartup].
      getService(Components.interfaces.nsIAppStartup);
    forceQuit  = Components.interfaces.nsIAppStartup.eForceQuit;

  }
  else if (kAppShell in Components.classes)
  {
    appService = Components.classes[kAppShell].
      getService(Components.interfaces.nsIAppShellService);
    forceQuit = Components.interfaces.nsIAppShellService.eForceQuit;
  }
  else
  {
    throw 'goQuitApplication: no AppStartup/appShell';
  }

  var windowManager = Components.
    classes['@mozilla.org/appshell/window-mediator;1'].getService();

  var windowManagerInterface = windowManager.
    QueryInterface(Components.interfaces.nsIWindowMediator);

  var enumerator = windowManagerInterface.getEnumerator(null);

  while (enumerator.hasMoreElements())
  {
    var domWindow = enumerator.getNext();
    if (("tryToClose" in domWindow) && !domWindow.tryToClose())
    {
      return false;
    }
    domWindow.close();
  }

  try
  {
    appService.quit(forceQuit);
  }
  catch(ex)
  {
    throw('goQuitApplication: ' + ex);
  }

  return true;
}



/***************
 * OTHER MODES *  
 ***************/


var win = (window.gSpider ? gSpider.mDocument.defaultView : window); // For spider.  I hope this is right (and safe).
var doc = win.document;


// Non-bookmarklet
//window.addEventListener("load", fuzzOnload, false); // !!!!

// Bookmarklet
// fuzzOnload();



/********************
 * COMMON DOM STUFF *
 ********************/
 
function addDOMNodes(n, enterIFrames, enterSVG, enterXBL) { 
  var i,c;

  if (n.tagName && n.tagName.toLowerCase().indexOf("script" != -1) && n.getAttribute("id") && n.getAttribute("id").indexOf("fuzz") != -1) {
    if (n.previousSibling && n.previousSibling.nodeType == Node.TEXT_NODE)
      alert("Warning: fuzz script found with a text (whitespace?) node before it.  This could affect node counts.");
    return;
  }

  if (n.tagName && n.tagName.toLowerCase() == "frameset") {
    gPageCompleted = true;
    throw "Avoiding bug 324318!";
  }
  

  // Add this node
  allNodes.push(n); 
  
  // Look for children
  for (i=0;c=n.childNodes[i];++i) {
    //try {
      addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
    //} catch(e) { /* Why would this fail? */ }
  }
  
  // If this is a frame, look inside it.
  if(enterIFrames) {
    try {
      if(n.contentDocument)
        try {
          addDOMNodes(n.contentDocument.documentElement, enterIFrames, enterSVG, enterXBL); 
        } catch(ex) { /* third-party iframe */ }
    } catch(ex) { /* avoid bug 310994 */ }
  }
  
  // If this is an <embed> referencing an SVG image, look inside it for SVG nodes.
  if (enterSVG) {
    if (n.getSVGDocument) { 
      try{
        if (n.getSVGDocument())
          addDOMNodes(n.getSVGDocument(), enterIFrames, enterSVG, enterXBL);
      } catch(ex) { }
    }
  }

  // Look for anonymous children (XBL, at least)
  // try/catch in case browser doesn't support getAnonymousNodes, or in case Firefox decides to throw a FAILURE error (!?).
  if (enterXBL) {
    try { 
      if (n.nodeType==1 && doc.getAnonymousNodes(n)) {
        for(i=0;c=doc.getAnonymousNodes(n)[i];++i) {
          if (!this.anonymousCount)
            anonymousCount = 0;
          ++anonymousCount;
          addDOMNodes(c, enterIFrames, enterSVG, enterXBL);
        }
      }
    } catch (ex) { } 
  }
}

// === Paste output from first run here, replacing the following two lines: ===

var fuzzSettings = null;
var fuzzCommands = null;

// === End pasteover ===


var fuzzerName = "XUL Element and Attribute fuzzer 3.0";
var recordStrategy = "Record as it goes";
var suggestedMaxSteps = 10000;

// dumpEachCommand = true;


var allNodes;
var fuzzRoot;

var XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

// dumpEachCommand = true;


function aC(q1, q2) { q1.appendChild(q2); }
function iB(q1, q2) { q1.parentNode.insertBefore(q2, q1); }
function rM(q1) { q1.parentNode.removeChild(q1); } // unused, but can be useful while narrowing testcases
// other things to try: clone instead of move, ...

function makeCommand()
{
  var n1index = rnd(allNodes.length), n2index = rnd(allNodes.length);
  
  var n1 = allNodes[n1index];
  var n2 = allNodes[n2index];
  
  var commandn1 = "allNodes[" + n1index + "]";
  var commandn2 = "allNodes[" + n2index + "]";
  var command;
  
  while(1){
  switch (rnd(9)) {
  case 0:
  case 1:
  case 2:
  
    // Change an attribute on n1, based on n1's tagName.
    
    if (!n1.tagName)
      break;
      
    var attrs = rnd(2) ? elements[n1.tagName] : commonAttributes;
    var attr = attrs ? rndElt[attrs] : null;
    if (!attr || rnd(5) == 1) {
      attr = totallyRandomAttr();
      // dumpln("*** " + attr);
    }
      
    var value = randomThing(rnd(5) == 1 ? attributes[totallyRandomAttr()] : attributes[attr]);
    
    if (attr) {
      if (rnd(2))
        // Set it as a property.
        return commandn1 + "['" + attr + "'] = " + simpleSource(value) + ";";
      else 
        // Set it as an attribute.
        return commandn1 + ".setAttribute('" + attr + "', " + simpleSource(value) + ");";
    }
      
    break;
    
  case 3:
    // Remove an attribute on n1, based on n1's current attribute list.
    
    var attrs = n1.attributes;
    if (attrs && attrs.length) {
      var attr = rndElt(attrs).name;
      return commandn1 + ".removeAttribute('" + attr + "');";
    }
    
    break;
    
  case 4:
    // Create a new element!
    // Set some attributes on it?  Nah.
    // Place it randomly, but only rarely place it outside of the document, because that's usually boring.
    
    var tag = rndElt(elementList);
    var newIndex = allNodes.length;
    

    var createness = "var ne = document.createElementNS(XUL_NS, '" + tag + "'); allNodes[" + newIndex + "] = ne; ";
    
    function isInDoc(n) { if (!n) return false; if (n.nodeType == Node.DOCUMENT_NODE) return true; return isInDoc(n.parentNode); }

    if (isInDoc(n1) != (rnd(10) == 1))
      return createness + commandn1 + ".appendChild(ne); ";
    else
      return createness + "fuzzRoot.appendChild(ne); ";

    break;

  case 5:
  case 6:
  case 7:
    // Move n2 in some way..
    
    if(n2index > 1) { // This fuzzer avoids screwing with the root node.
  
      if (rnd(20) == 1)
        // Don't do this too often -- it tends to make things less interesting.
        return "rM(" + commandn2 + ");";
      
      if (rnd(9) == 3)
        // Don't do this too often either -- it hurts reduction and it doesn't buy us much
        return "iB(" + commandn1 + ", " + commandn2 + ");";
        
      // Do this a lot ;)
      return "aC(" + commandn1 + ", " + commandn2 + ");";
    }
    
    break;
    
  case 8:
    // Add some text -- this is partly for visual effect rather than finding more bugs.
  
    switch(rnd(3)) {
    case 0:
      return "aC(" + commandn1 + ", allNodes[" + allNodes.length + "] = document.createTextNode('<" + n1.tagName + ">'));";
    case 1:
      return commandn1 + ".setAttribute('value', '`" + n1.tagName + "`');";
    case 2:
      return commandn1 + ".setAttribute('label', '/" + n1.tagName + "/');";
    }
  
  }
  }
  
}

function doCommand(x) { eval(x); }



function resetNodeList() {
  allNodes = [];
  fuzzRoot = doc.body || doc.documentElement;
  addDOMNodes(fuzzRoot);
}


function initFuzzerSpecific() { 
  resetNodeList();
  startFuzzing();
}



function randomThing(obj)
{
  if (obj == null)
    return null;
    
  if (typeof obj == "function")
    return obj();
    
  if (typeof obj == "string")
    return obj; // not to be treated as an array :)

  if (obj.length) { // non-empty array
    var x = rndElt(obj);
    //dumpln(x);
    return randomThing(x)
  }

  if (obj.length === 0)
    return null;

  // Otherwise, it should be a number, etc.
  return obj;
}



var names =    [  "a",  "b",  "c",  "d",  "e",  "f",  "g",  "1" ];
var chars = names;
var namerefs = [ "#a", "#b", "#c", "#d", "#e", "#f", "#g", "#1" ];
var charsets = [ "UTF-8", "ISO-8859-1" ];
var URIs = [
  ["http://www.squarefree.com/", "http://www.squarefree.com/feed"], 
  ["#a", "#b"], 
  "data:text/html,", 
  ["javascript:alert(Components.classes)", "javascript:5555", "javascript:'QQQQ' + String.fromCharCode(0) + 'UUUU'"], 
  "http://www.mozilla.org/status/tree.gif"
  ];
  
  // gopher://seanm.ca/
  
var numbers = [-2, -1, 0, "-0", "+0", 0.5, 1, "+1", 2, 3, 4, 5, 6, 7, 8, 100];
function numbersWithUnits() { return rndElt(numbers) + rndElt(["px", "em", "pt", "%", ""]) }
var relations = ["alternate"];
var mimeTypes = ["text/html", "text/plain", "text/css", "text/javascript", "application/rss+xml"];
var texts = ["", "foo", "bar", "baz", "zap", "Abbey\nRoad", 'QQQQ' + String.fromCharCode(0) + 'UUUU'];
var colors = ["lime", "red", "blue", "#CCC", "#777777"];
var percents = ["0%", "-10%", "50%", "80%", "99.999%", "100%", "120%"];
var scriptcode = [// "alert('script')", 
  ""
// Neutered to avoid hitting bug 353610
//  "var t = event.originalTarget; if(t) { dumpln('script 1 called! ' + t.tagName + ' ' + event.type); t.parentNode.removeChild(t); }",
//  "var t = event.originalTarget; if(t) { dumpln('script 2 called! ' + t.tagName + ' ' + event.type); var p = t.parentNode; p.parentNode.removeChild(p); }"
// Ideas for things to add
//  "throw 1"
//  "document.removeChild(document.documentElement);" // and put it back later?
  //"window.close()"
  ];
var booleans = ["true", "false"];

function longThing()
{
  // a string of all A's in a buffer overflow has the nice property
  // of a guaranteed crash on some architectures, IIRC
  
  var x = "A"; 
  for (var i = 0; i < 14; ++i)
    x = x + x;
  return x; 
}


// http://xulplanet.com/references/elemref/ref_XULElement.html#attr_align
var commonAttributes = ["align", "allowevents", "allownegativeassertions", "class", "coalesceduplicatearcs", 
"collapsed", "container", "containment", "context", "contextmenu",
"datasources", "dir", "empty", "equalsize", "flags",
"flex", "height", "hidden", "id", "insertafter",
"insertbefore", "left", "maxheight", "maxwidth", "menu",
"minheight", "minwidth", "mousethrough", "observes", "ordinal", 
"orient", "pack", "persist", "popup", "position",
"preference-editable", "ref", "removeelement", "sortDirection", "sortResource",
"sortResource2", "statustext", "style", "template", "tooltip",
"tooltiptext",  "top", "uri", "wait-cursor", "width"];


var attributes = {
"randomcrap": ["", "\0", longThing],


// From http://www.xulplanet.com/references/elemref/ref_XULElement.html
// Quite a few of these only apply to overlays, or to RDF-generated stuff.

"id": names,
"name": names,
"align": ["start", "center", "end", "baseline", "stretch", "left", "right"],
"allowevents": booleans,
"allownegativeassertions": booleans,
"class": names,
"coalesceduplicatearcs": booleans,
"collapsed": booleans,
"container": booleans,
"containment": URIs, // Type: URI of an RDF predicate
"context": names, // Type: popup element id.
"contextmenu": names,
"datasources": URIs, // Actually multiple URIs
"dir": [
  ["normal", "reverse"], // most elements
  ["left", "right", "top", "bottom", "bottomleft", "bottomright", "topleft", "topright"] // resizer
  ],
"empty": booleans,
"equalsize": ["always", "never"],
"flags": [
  ["dont-test-empty", "dont-build-content", "dont-test-empty dont-build-content"], // all elements
  "dont-build-content" // tree
  ],
"flex": numbers,
"height": numbers,
"hidden": booleans,
"id": names,
"insertafter": names,
"insertbefore": names,
"left": numbers,
"maxheight": numbers,
"maxwidth": numbers,
"menu": names, // Type: popup element id
"minheight": numbers,
"minwidth": numbers,
"mousethrough": ["always", "never"],
"observes": names, // Type: broadcaster element id
"ordinal": numbers,
"orient": ["horizontal", "vertical"],
"pack": ["start", "center", "end"],
// "persist": ??, // Type: space separated list of attribute names
"popup": names, // Type: popup element id
"position": numbers, // index into a child list
"preference-editable": booleans,
"ref": URIs, // Type: URI of an RDF resource
"removeelement": names,
"sortDirection": ["ascending", "descending", "natural"],
"sortResource": URIs,
"sortResource2": URIs,
"statustext": texts,
// "style": ??
"template": names, // ??

"tooltip": names,
"tooltiptext": texts,

"top": numbers, // for elements in a stack??
"uri": ["rdf:*", URIs], // ? (special meaning for <xul:content>)
"wait-cursor": booleans, // ?
"width": numbers,

// overloaded

"type": [
  "content", "content-primary", // browser
  "button", // colorpicker
  "checkbox", // listitem, menuitem
  "radio", // menuitem
  ["bool", "int", "string", "unichar", "wstring", "file"], // preference
  ["autocomplete", "password", "timed"], // textbox
  ["menu", "menu-button"], // toolbarbutton  
  ["checkbox", "progressmeter", "text"], // treecol
  ],

// <xul:browser>
"autocompleteenabled": booleans,
"autocompletepopup": names,
"autoscroll": booleans,
"disablehistory": booleans,
"disablesecurity": booleans,
"homepage": URIs,
"src": URIs,

// checkbox
"accesskey": chars,
"checked": booleans,
"command": names,
"crop": ["start", "end", "left", "right", "center", "none"],
"disabled": booleans,
"image": URIs,
"label": texts,
"preference": names,
"tabindex": numbers,

// colorpicker
"color": colors,

// commandset
"commandupdater": booleans,
"events": ["focus", "select", "clipboard", "undo", "*"],
"oncommandupdate": scriptcode,
"targets": names, // really multiple names

// command
"oncommand": scriptcode,

// content
"tag": ["treechildren", "content"], // ?

// deck
"selectedIndex": numbers,

// description
"control": names,
"value": [texts, percents], // percents for progressmeter

// dialogheader
"title": texts,
"description": texts,

// editor
"editortype": ["html", "text"],

// image
"onerror": scriptcode,
"onload": scriptcode,
"src": URIs,
"validate": ["always", "never"], // also menuitem

// key
"key": [chars, names], // chars in <xul:key>, names in e.g. <xul:menu> referencing a <xul:key>.
"keycode": ["VK_RETURN", "VK_ENTER", "VK_I"],
"modifiers": ["shift", "alt", "meta", "control", "accel", "accel,shift"],

// listbox
"disableKeyNavigation": booleans,
"rows": numbers,
"seltype": ["single", "multiple"],
"suppressonselect": booleans,
"onselect": scriptcode,

// listitem
"current": booleans,

// menu
"acceltext": [chars, texts],
"allowevents": booleans,
"menuactive": booleans,
"open": booleans,
"sizetopopup": ["none", "always"], // not never/always!? according to http://xulplanet.com/references/elemref/ref_menu.html

// menubar
"grippyhidden": booleans,
"statusbar": names,

// menuitem
"autocheck": booleans,

// menulist
"readonly": booleans,
"focused": booleans, // !!
"disableautoselect": booleans,
"editable": booleans,

// menupopup
"ignorekeys": booleans,
"top": numbers,
"left": numbers,
"onpopuphidden": scriptcode,
"onpopuphiding": scriptcode,
"onpopupshowing": scriptcode,
"onpopupshown": scriptcode,
"position": ["after_start", "after_end", "before_start", "before_end", "end_after", "end_before", "start_after", "start_before", "overlap", "at_pointer", "after_pointer"],

// observes
"attribute": ["*", "attribute", "", "disabled"],
"element": names,

// preference
"inverted": booleans,
"instantApply": booleans,

// prefpane
"helpURI": URIs,
"onpaneload": scriptcode,


"mode": [
  ["determined", "undetermined"], // progressmeter
  ["none", "normal", "undetermined"] // treecell
  ],

// scrollbar
"curpos": numbers, 
"increment": numbers, 
"maxpos": numbers, 
"pageincrement": numbers,

// splitter
"collapse": ["none", "before", "after"],
"resizeafter": ["closest", "farthest", "grow"],
"resizebefore": ["closest", "farthest"],
"state": ["open", "collapsed", "dragging"],

// tab
"afterselected": booleans,
"beforeselected": booleans,
"linkedpanel": names,

// tabbrowser
"autoscroll": booleans,
"contentcontextmenu": names,
"contenttooltip": names,
"handleCtrlPageUpDown": booleans,
"onbookmarkgroup": scriptcode,
"onnewtab": scriptcode,

// tabbox
"handleCtrlTab": booleans,
"eventnode": ["parent", "window", "document"],

// tabs
"closebutton": booleans,
"disableclose": booleans,
"first-tab": booleans,
"last-tab": booleans,
"onclosetab": scriptcode,
"onnewtab": scriptcode,
"onselect": scriptcode,
"setfocus": booleans,
"tooltiptextnew": texts,

// textbox
"cols": numbers,
"maxlength": numbers,
"multiline": booleans,
"onchange": scriptcode,
"oninput": scriptcode,
"timeout": numbers,
"wrap": ["off"],

// toolbar
"currentset": names, // :(
"customindex": numbers,
"customizable": booleans,
"defaultset": names,
"grippyhidden": booleans,
"grippytooltiptext": texts,
"toolbarname": texts,

// toolbarbutton
"autoCheck": booleans,
"checkState": [0,1,2],
"dlgType": ["accept", "cancel", "help", "disclosure"],
"group": names, // ?

// tooltip
"default": booleans,
"noautohide": booleans,

// tree
"enableColumnDrag": booleans,
"hidecolumnpicker": booleans,
"statedatasource": URIs,

// treechildren
"alternatingbackground": booleans,

// treecol
"cycler": booleans,
"dragging": booleans,
"fixed": booleans,
"hidden": booleans,
"hideheader": booleans,
"ignoreincolumnpicker": booleans,
"primary": booleans,
"sortActive": booleans,

// treeitem
"open": booleans,

// window
"hidechrome": booleans,
"screenX": numbers,
"screenY": numbers,
"sizemode": ["maximized", "minimized", "normal"],
"windowtype": names,

// causing trouble?
//"onmousemove": scriptcode,


};




var elements = 
{
"action": [],
"arrowscrollbox": ["orient"],
"bbox": [],
"binding": ["object", "predicate", "subject"],
"bindings": [],
"box": [],
"broadcaster": ["id"],
"broadcasterset": [],
"button": ["label"],
// "browser": ["autocompleteenabled", "autocompletepopup", "autoscroll", "disablehistory", "disablesecurity", "homepage", "src", "type"],
  // browser sounds dangerous ;)
  // and indeed it is. commented out to avoid hitting bug 344211.
  
"checkbox": ["accesskey", "checked", "command", "crop", "disabled", "image", "label", "preference", "tabindex"],
"caption": ["accesskey", "crop", "image", "label", "tabindex"],
"colorpicker": ["color", "preference", "tabindex", "type"],
"column": [],
"columns": [],
"commandset": ["commandupdater", "events", "oncommandupdate", "targets"],
"command": ["oncommand"],
"conditions": [],
"content": ["tag", "uri"],
"deck": ["selectedIndex"],
"description": ["accesskey", "control", "crop", "disabled", "value"],
"dialog": [], // has tons of attributes; skipping them.
"dialogheader": ["crop", "description", "title"],

// editor seems to cause the same problems as iframes (bug 344216)
// "editor": ["editortype", "src", "type"],

"grid": [],
"grippy": [],
"groupbox": [],
"hbox": [],

// iframe commented out to avoid triggering bug 344216 and maybe speed things up.
// "iframe": ["src"],


"image": ["onerror", "onload", "src", "validate"],
"key": ["command", "disabled", "key", "keycode",  "keytext", "modifiers", "oncommand", "phase"],
"keyset": [],
"label": ["accesskey", "control", "crop", "disabled", "value"],
// listboxes are too crashy, and i don't feel like filing more bugs on them before bug 321016 is fixed
// "listbox": ["disableKeyNavigation", "disabled", "preference", "rows", "seltype", "suppressonselect", "tabindex", "value"],
// should "listboxbody" be here, once bug 321016 is fixed?
"listcell": ["crop", "disabled", "image", "label", "type"],
"listcol": [],
"listcols": [],
"listhead": [],
"listheader": ["disabled"],
"listitem": ["accesskey", "checked", "command", "crop", "current", "disabled", "image", "label", "preference", "selected", "tabindex", "type", "value"],
"member": ["child", "container"], // advanced rdf rules stuff
"menu": ["acceltext", "accesskey", "allowevents", "crop", "disabled", "key", "label", "menuactive", "open", "sizetopopup", "value"],
"menubar": ["grippyhidden", "statusbar"],
"menuitem": ["acceltext", "accessible", "accesskey", "allowevents", "autocheck", "checked", "command", "crop", "description", "image", "key", "label", "name", "selected", "tabindex", "type", "validate", "value"],
"menulist": ["accesskey", "crop", "disableautoselect", "disabled", "editable", "focused", "image", "label", "open", "preference", "readonly", "sizetopopup", "src", "tabindex", "value"],
"menupopup": ["ignorekeys", "left", "onpopuphidden", "onpopuphiding", "onpopupshowing", "onpopupshown", "position", "top"],
"menuseparator": [],

// Observes commented out to avoid triggering bug 344215.
// "observes": ["attribute", "element"],

"overlay": ["class", "id"],
"page": [], // has tons of attributes; skipping them.
"popup": ["ignorekeys", "left", "onpopuphidden", "onpopuphiding", "onpopupshowing", "onpopupshown", "position", "top"],
"popupset": [],

// Preferences commented out to avoid triggering bug 344334.
/*
"preference": ["disabled", "instantApply", "inverted", "name", "onchange", "readonly", "tabindex", "type"],
"preferences": [],
*/

"prefpane": ["helpURI", "image", "label", "onpaneload", "selected", "src"],
"prefwindow": [], // has tons of attributes; skipping them.
"progressmeter": ["mode", "value"],
"radio": ["accesskey", "command", "crop", "disabled", "focused", "image", "label", "selected", "tabindex", "value"],
"radiogroup": ["disabled", "focused", "preference", "tabindex", "value"],
"resizer": ["dir"],

// Richlistbox commented out to avoid triggering bug 344228.
/*
"richlistbox": ["suppressonselect"],
"richlistitem": ["disabled", "selected", "tabindex", "value"],
*/

"row": [],
"rows": [],
"rule": ["iscontainer", "isempty", "parent", "parsetype"],
"script": ["src", "type"],
"scrollbar": ["curpos", "increment", "maxpos", "pageincrement"],
"scrollbox": [],
"scrollcorner": [],
"separator": ["orient"],
"spacer": ["flex"],
"splitter": ["collapse", "resizeafter", "resizebefore", "state"],
"stack": [],
"statusbar": [],
"statusbarpanel": ["crop", "label", "src"],
"stringbundle": ["src"],
"stringbundleset": [],
"tab": ["accesskey", "afterselected", "beforeselected", "crop", "disabled", "image", "label", "linkedpanel", "selected", "validate"],
// "tabbrowser": ["autocompleteenabled", "autocompletepopup", "autoscroll", "contentcontextmenu", "contenttooltip", "handleCtrlPageUpDown", "onbookmarkgroup", "onnewtab"], // commented out for the same reason as <browser>
"tabbox": ["eventnode", "handleCtrlPageUpDown", "handleCtrlTab"],
"tabpanel": [],
"tabpanels": ["selectedIndex"],
"tabs": ["closebutton", "disableclose", "first-tab", "last-tab", "onclosetab", "onnewtab", "onselect", "setfocus", "tooltiptextnew"],
"template": ["container", "member"],
"textnode": ["value"],
"textbox": ["cols", "disabled", "maxlength", "multiline", "onchange", "oninput", "preference", "readonly", "rows", "size", "tabindex", "timeout", "type", "value", "wrap"],
"titlebar": [],

// Toolbar stuff commented out to avoid triggering bug 344434.
/*
"toolbar": ["currentset", "customindex", "customizable", "defaultset", "grippyhidden", "grippytooliptext", "toolbarname"],
"toolbarbutton": ["accesskey", "autoCheck", "checkState", "checked", "command", "crop", "dir", "disabled", "dlgType", "group", "image", "label", "open", "orient", "tabindex", "type", "validate"],
"toolbargrippy": [],
"toolbaritem": [],
"toolbarpalette": [],
"toolbarseparator": [],
"toolbarset": [],
"toolbarspaer": [],
// "toolbarspring": [], // toolbarspring also commented out to avoid triggering bug 344340
*/

"toolbox": [],

// Tooltip commented out to avoid triggering bug 344340.
/*
"tooltip": ["crop", "default", "label", "noautohide", "onpopuphidden", "onpopuphiding", "onpopupshowing", "onpopupshown", "position"], 
*/

// Trees commented out to avoid triggering bug 344228.
/*
"tree": ["disableKeyNavigation", "disabled", "enableColumnDrag", "flags", "hidecolumnpicker", "onselect", "rows", "seltype", "statedatasource", "tabindex"],
"treecell": ["label", "mode", "properties", "ref", "src", "value"],
"treechildren": ["alternatingbackground"],
"treecol": ["crop", "cycler", "dragging", "fixed", "hidden", "hideheader", "ignoreincolumnpicker", "label", "primary", "sort", "sortActive", "sortDirection", "src", "type"],
"treecols": [],
"treeitem": ["container", "empty", "label", "open", "uri"],
"treerow": ["properties"],
"treeseparator": ["properties"],
*/

"triple": ["object", "predicate", "subject"],
"vbox": [],
"window": ["height", "hidechrome", "id", "screenX", "screenY", "sizemode", "title", "width", "windowtype"],
"wizard": [], // ...
"wizardpage": ["next", "pageid"]
};




// attributes is a hash, but sometimes we want to treat it as an array-like thing.  so....
attributeList = [];
for (var prop in attributes)
  attributeList.push(prop);

function totallyRandomAttr() { 
  // Emphasize basic stuff. 
  if (rnd(3)==1) 
    return rndElt(["id", "name"]);
  return rndElt(attributeList); 
}
// attributes is a hash, but sometimes we want to treat it as an array-like thing.  so....
elementList = [];
for (var prop in elements)
  elementList.push(prop);

fuzzOnload();